# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for should_disable_afc."""

import pytest
from ... import types
from ..._extra_utils import should_disable_afc


def test_config_is_none():
  assert should_disable_afc(None) is False


def test_afc_config_unset():
  assert should_disable_afc(types.GenerateContentConfig()) is False


def test_afc_enable_unset_max_0():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  maximum_remote_calls=0,
              ),
          )
      )
      is True
  )


def test_afc_enable_unset_max_negative():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  maximum_remote_calls=-1,
              ),
          )
      )
      is True
  )


def test_afc_enable_unset_max_0_0():
  assert (
      should_disable_afc(
          {'automatic_function_calling': {'maximum_remote_calls': 0.0}}
      )
      is True
  )


def test_afc_enable_unset_max_1():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  maximum_remote_calls=1,
              ),
          )
      )
      is False
  )


def test_afc_enable_unset_max_1_0():
  assert (
      should_disable_afc(
          {'automatic_function_calling': {'maximum_remote_calls': 1.0}}
      )
      is False
  )


def test_afc_enable_false_max_unset():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=True,
              ),
          )
      )
      is True
  )


def test_afc_enable_false_max_0():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=True,
                  maximum_remote_calls=0,
              ),
          )
      )
      is True
  )


def test_afc_enable_false_max_negative():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=True,
                  maximum_remote_calls=-1,
              ),
          )
      )
      is True
  )


def test_afc_enable_false_max_0_0():
  assert (
      should_disable_afc(
          {'automatic_function_calling': {'maximum_remote_calls': 0.0}}
      )
      is True
  )


def test_afc_enable_false_max_1():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=True,
                  maximum_remote_calls=1,
              ),
          )
      )
      is True
  )


def test_afc_enable_true_max_unset():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=False,
              ),
          )
      )
      is False
  )


def test_afc_enable_true_max_0():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=False,
                  maximum_remote_calls=0,
              ),
          )
      )
      is True
  )


def test_afc_enable_true_max_negative():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=False,
                  maximum_remote_calls=-1,
              ),
          )
      )
      is True
  )


def test_afc_enable_true_max_0_0():
  assert (
      should_disable_afc(
          {'automatic_function_calling': {'maximum_remote_calls': 0.0}}
      )
      is True
  )


def test_afc_enable_true_max_1():
  assert (
      should_disable_afc(
          types.GenerateContentConfig(
              automatic_function_calling=types.AutomaticFunctionCallingConfig(
                  disable=False,
                  maximum_remote_calls=1,
              ),
          )
      )
      is False
  )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for tunings.tune()."""

from ... import types as genai_types
from .. import pytest_helper
import pytest


VERTEX_HTTP_OPTIONS = {
    'api_version': 'v1beta1',
    'base_url': 'https://us-central1-autopush-aiplatform.sandbox.googleapis.com/',
}

evaluation_config=genai_types.EvaluationConfig(
    metrics=[
        genai_types.Metric(name="bleu", prompt_template="test prompt template")
    ],
    output_config=genai_types.OutputConfig(
        gcs_destination=genai_types.GcsDestination(
            output_uri_prefix="gs://sararob_test/"
        )
    ),
)

test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name="test_dataset_gcs_uri",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
            ),
        ),
        exception_if_mldev="gcs_uri parameter is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_pretuned_model",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="projects/801452371447/locations/us-central1/models/9030969596621881344",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_train_data.jsonl",
            ),
        ),
        exception_if_mldev="is not supported in Gemini API",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_pretuned_model_with_checkpoint_id",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="projects/801452371447/locations/us-central1/models/9030969596621881344",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_train_data.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                pre_tuned_model_checkpoint_id="3",
            ),
        ),
        exception_if_mldev="is not supported in Gemini API",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_simple_dpo",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                tuned_model_display_name="Model display name",
                epoch_count=1,
                method="PREFERENCE_TUNING",
            ),
        ),
        exception_if_mldev="parameter is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_dpo_with_beta",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                tuned_model_display_name="Model display name",
                epoch_count=1,
                method=genai_types.TuningMethod.PREFERENCE_TUNING,
                beta=0.5,
            ),
        ),
        exception_if_mldev="parameter is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_non_pretuned_model_with_checkpoint_id",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                pre_tuned_model_checkpoint_id="3",
            ),
        ),
        exception_if_mldev="is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_dataset_gcs_uri_all_parameters",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                tuned_model_display_name="Model display name",
                epoch_count=1,
                learning_rate_multiplier=1.0,
                adapter_size="ADAPTER_SIZE_ONE",
                validation_dataset=genai_types.TuningValidationDataset(
                    gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_validation_data.jsonl",
                ),
            ),
        ),
        exception_if_mldev="gcs_uri parameter is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_dataset_vertex_dataset_resource",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                vertex_dataset_resource="projects/801452371447/locations/us-central1/datasets/5779918772206829568",
            ),
        ),
        exception_if_mldev="vertex_dataset_resource parameter is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_dataset_dataset_resource_all_parameters",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                vertex_dataset_resource="projects/801452371447/locations/us-central1/datasets/5779918772206829568",
            ),
            config=genai_types.CreateTuningJobConfig(
                tuned_model_display_name="Model display name",
                epoch_count=1,
                learning_rate_multiplier=1.0,
                adapter_size="ADAPTER_SIZE_ONE",
                validation_dataset=genai_types.TuningValidationDataset(
                    vertex_dataset_resource="projects/801452371447/locations/us-central1/datasets/1168232753779441664",
                ),
                labels={"testlabelkey": "testlabelvalue"},
            ),
        ),
        exception_if_mldev="vertex_dataset_resource parameter is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_distillation",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="meta/llama3_1@llama-3.1-8b-instruct",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-openai-opposites.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                method="DISTILLATION",
                base_teacher_model="deepseek-ai/deepseek-v3.1-maas",
                epoch_count=20,
                validation_dataset=genai_types.TuningValidationDataset(
                    gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-val-openai-opposites.jsonl",
                ),
                output_uri="gs://nathreya-oss-tuning-sdk-test/ayushagra-distillation-test-folder",
                http_options=VERTEX_HTTP_OPTIONS,
            ),
        ),
        exception_if_mldev="parameter is not supported in Gemini API.",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_oss_sft",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="meta/llama3_1@llama-3.1-8b-instruct",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-openai-opposites.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                epoch_count=20,
                validation_dataset=genai_types.TuningValidationDataset(
                    gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-val-openai-opposites.jsonl",
                ),
                custom_base_model="gs://nathreya-oss-tuning-sdk-test/ayushagra-distillation-test-folder/postprocess/node-0/checkpoints/final",
                output_uri="gs://nathreya-oss-tuning-sdk-test/ayushagra-distillation-test",
                http_options=VERTEX_HTTP_OPTIONS,
            ),
        ),
        exception_if_mldev="not supported in Gemini API",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_oss_sft_hyperparams",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="meta/llama3_1@llama-3.1-8b-instruct",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-openai-opposites.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                epoch_count=20,
                validation_dataset=genai_types.TuningValidationDataset(
                    gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-val-openai-opposites.jsonl",
                ),
                learning_rate=2.5e-4,
                tuning_mode="TUNING_MODE_FULL",
                custom_base_model="gs://nathreya-oss-tuning-sdk-test/ayushagra-distillation-test-folder/postprocess/node-0/checkpoints/final",
                output_uri="gs://nathreya-oss-tuning-sdk-test/ayushagra-distillation-test",
                http_options=VERTEX_HTTP_OPTIONS,
            ),
        ),
        exception_if_mldev="not supported in Gemini API",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_oss_distillation",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="meta/llama3_1@llama-3.1-8b-instruct",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-openai-opposites.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                method="DISTILLATION",
                base_teacher_model="deepseek-ai/deepseek-v3.1-maas",
                epoch_count=20,
                validation_dataset=genai_types.TuningValidationDataset(
                    gcs_uri="gs://nathreya-oss-tuning-sdk-test/distillation-val-openai-opposites.jsonl",
                ),
                custom_base_model="gs://nathreya-oss-tuning-sdk-test/ayushagra-distillation-test-folder/postprocess/node-0/checkpoints/final",
                output_uri="gs://nathreya-oss-tuning-sdk-test/ayushagra-distillation-test",
                http_options=VERTEX_HTTP_OPTIONS,
            ),
        ),
        exception_if_mldev="not supported in Gemini API",
    ),
    pytest_helper.TestTableItem(
        name="test_tune_encryption_spec",
        parameters=genai_types.CreateTuningJobParameters(
            base_model="gemini-2.5-flash",
            training_dataset=genai_types.TuningDataset(
                gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
            ),
            config=genai_types.CreateTuningJobConfig(
                encryption_spec=genai_types.EncryptionSpec(
                    kms_key_name="projects/vertex-sdk-dev/locations/us-central1/keyRings/ayush-test-key/cryptoKeys/ayush-key",
                ),
            ),
        ),
        exception_if_mldev="not supported in Gemini API",
    ),
]

pytestmark = pytest_helper.setup(
    file=__file__,
    globals_for_file=globals(),
    test_method="tunings.tune",
    test_table=test_table,
)

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.skipif(
    "config.getoption('--private')",
    reason="Skipping in pre-public tests"
)
def test_eval_config(client):
  """Tests tuning with eval config."""

  if client._api_client.vertexai:
    tuning_job = client.tunings.tune(
      base_model="gemini-2.5-flash",
      training_dataset=genai_types.TuningDataset(gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_train_data.jsonl"),
      config=genai_types.CreateTuningJobConfig(
          tuned_model_display_name="tuning job with eval config",
          epoch_count=1,
          learning_rate_multiplier=1.0,
          adapter_size="ADAPTER_SIZE_ONE",
          validation_dataset=genai_types.TuningValidationDataset(
              gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_validation_data.jsonl"
          ),
          evaluation_config=evaluation_config,
      ),
    )

    assert tuning_job.state == genai_types.JobState.JOB_STATE_PENDING
    assert tuning_job.evaluation_config == evaluation_config


@pytest.mark.skipif(
    "config.getoption('--private')",
    reason="Skipping in pre-public tests"
)
def test_eval_config_with_metrics(client):
  """Tests tuning with eval config metrics."""
  if client._api_client.vertexai:
    evaluation_config=genai_types.EvaluationConfig(
        metrics=[
            genai_types.Metric(
                name="prompt-relevance",
                prompt_template="How well does the response address the prompt?: PROMPT: {request}\n RESPONSE: {response}\n",
                return_raw_output=True,
                judge_model_system_instruction="You are a cat. Make all evaluations from this perspective.",
            ),
            genai_types.Metric(name='bleu'),
            genai_types.Metric(name='rouge_1'),
        ],
        output_config=genai_types.OutputConfig(
            gcs_destination=genai_types.GcsDestination(
                output_uri_prefix="gs://sararob_test/"
            )
        ),
        autorater_config=genai_types.AutoraterConfig(
            sampling_count=1,
            autorater_model="test-model",
        ),
    )
    tuning_job = client.tunings.tune(
      base_model="gemini-2.5-flash",
      training_dataset=genai_types.TuningDataset(gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_train_data.jsonl"),
      config=genai_types.CreateTuningJobConfig(
          tuned_model_display_name="tuning job with eval config",
          epoch_count=1,
          learning_rate_multiplier=1.0,
          adapter_size="ADAPTER_SIZE_ONE",
          validation_dataset=genai_types.TuningValidationDataset(
              gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_validation_data.jsonl"
          ),
          evaluation_config=evaluation_config,
      ),
    )
    assert tuning_job.state == genai_types.JobState.JOB_STATE_PENDING


@pytest.mark.skipif(
    "config.getoption('--private')",
    reason="Skipping in pre-public tests"
)
def test_eval_config_with_unified_metrics(client):
  """Tests tuning with eval config metrics."""
  if client._api_client.vertexai:
    evaluation_config=genai_types.EvaluationConfig(
        metrics=[
            genai_types.UnifiedMetric(
                pointwise_metric_spec=genai_types.PointwiseMetricSpec(
                  metric_prompt_template=(
                      "How well does the response address the prompt?: "
                      "PROMPT: {request}\n RESPONSE: {response}\n"
                  ),
                  system_instruction=(
                      "You are a cat. Make all evaluations from this perspective."
                  ),
                  custom_output_format_config=genai_types.CustomOutputFormatConfig(
                      return_raw_output=True
                  ),
                )
            ),
            genai_types.UnifiedMetric(
                bleu_spec=genai_types.BleuSpec(use_effective_order=True)
            ),
            genai_types.UnifiedMetric(
                rouge_spec=genai_types.RougeSpec(rouge_type="rouge1")
            ),
        ],
        output_config=genai_types.OutputConfig(
            gcs_destination=genai_types.GcsDestination(
                output_uri_prefix="gs://sararob_test/"
            )
        ),
        autorater_config=genai_types.AutoraterConfig(
            sampling_count=1,
            autorater_model="test-model",
        ),
    )
    tuning_job = client.tunings.tune(
      base_model="gemini-2.5-flash",
      training_dataset=genai_types.TuningDataset(gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_train_data.jsonl"),
      config=genai_types.CreateTuningJobConfig(
          tuned_model_display_name="tuning job with eval config",
          epoch_count=1,
          learning_rate_multiplier=1.0,
          adapter_size="ADAPTER_SIZE_ONE",
          validation_dataset=genai_types.TuningValidationDataset(
              gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_validation_data.jsonl"
          ),
          evaluation_config=evaluation_config,
      ),
    )
    assert tuning_job.state == genai_types.JobState.JOB_STATE_PENDING


@pytest.mark.skipif(
    "config.getoption('--private')",
    reason="Skipping in pre-public tests"
)
def test_eval_config_with_metrics_dict(client):
  """Tests tuning with eval config metrics."""
  if client._api_client.vertexai:
    evaluation_config=genai_types.EvaluationConfig(
        metrics = [
            {
                "name": "prompt-relevance",
                "prompt_template": "How well does the response address the prompt?: PROMPT: {request}\n RESPONSE: {response}\n",
                "return_raw_output": True,
                "judge_model_system_instruction": "You are a cat. Make all evaluations from this perspective.",
            },
            {"name": "bleu"},
            {"name": "rouge_1"},
            {
                "bleu_spec": {
                    "use_effective_order": True
                }
            },
        ],
        output_config=genai_types.OutputConfig(
            gcs_destination=genai_types.GcsDestination(
                output_uri_prefix="gs://sararob_test/"
            )
        ),
        autorater_config=genai_types.AutoraterConfig(
            sampling_count=1,
            autorater_model="test-model",
        ),
    )
    tuning_job = client.tunings.tune(
      base_model="gemini-2.5-flash",
      training_dataset=genai_types.TuningDataset(gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_train_data.jsonl"),
      config=genai_types.CreateTuningJobConfig(
          tuned_model_display_name="tuning job with eval config",
          epoch_count=1,
          learning_rate_multiplier=1.0,
          adapter_size="ADAPTER_SIZE_ONE",
          validation_dataset=genai_types.TuningValidationDataset(
              gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_validation_data.jsonl"
          ),
          evaluation_config=evaluation_config,
      ),
    )
    assert tuning_job.state == genai_types.JobState.JOB_STATE_PENDING



@pytest.mark.skipif(
    "config.getoption('--private')",
    reason="Skipping in pre-public tests"
)
@pytest.mark.asyncio
async def test_eval_config_async(client):
  """Tests tuning with eval config."""
  if client._api_client.vertexai:
    tuning_job = await client.aio.tunings.tune(
      base_model="gemini-2.5-flash",
      training_dataset=genai_types.TuningDataset(gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_train_data.jsonl"),
      config=genai_types.CreateTuningJobConfig(
          tuned_model_display_name="tuning job with eval config",
          epoch_count=1,
          learning_rate_multiplier=1.0,
          adapter_size="ADAPTER_SIZE_ONE",
          validation_dataset=genai_types.TuningValidationDataset(
              gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-2_0/text/sft_validation_data.jsonl"
          ),
          evaluation_config=evaluation_config,
      ),
    )

    assert tuning_job.state == genai_types.JobState.JOB_STATE_PENDING



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for caches.create(), caches.get(), and caches.delete()."""

from .... import types as genai_types
from ... import pytest_helper


def create_get_delete(client, parameters):
  if client.vertexai:
    cache = client.caches.create(
        model=parameters.model,
        config=parameters.config,
    )
  else:
    # For MLDev, we must create the file before creating the cache.
    file = client.files.upload(file="tests/data/a-man-and-a-dog.png")
    config = genai_types.CreateCachedContentConfig(
        contents=[
            genai_types.Part.from_uri(
                file_uri=file.uri,
                mime_type=file.mime_type,
            )
        ]
        * 5
    )
    cache = client.caches.create(model=parameters.model, config=config)
  got_cache = client.caches.get(name=cache.name)
  client.caches.delete(name=got_cache.name)


test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name="test_create_get_delete",
        parameters=genai_types._CreateCachedContentParameters(
            model="gemini-2.5-flash",
            config=genai_types.CreateCachedContentConfig(
                contents=[genai_types.Part.from_uri(
                    file_uri="gs://cloud-samples-data/generative-ai/image/a-man-and-a-dog.png",
                    mime_type="image/png",
                )] * 5
            ),
        ),
    ),
]

pytestmark = pytest_helper.setup(
    file=__file__,
    globals_for_file=globals(),
    test_method="create_get_delete",
    test_table=test_table,
)

pytest_plugins = ("pytest_asyncio",)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable
from typing_extensions import Literal, Required, TypeAlias, TypedDict

from .text_content_param import TextContentParam
from .image_content_param import ImageContentParam

__all__ = ["FunctionResultContentParam", "Result", "ResultItems", "ResultItemsItem"]

ResultItemsItem: TypeAlias = Union[TextContentParam, ImageContentParam]


class ResultItems(TypedDict, total=False):
    items: Iterable[ResultItemsItem]


Result: TypeAlias = Union[ResultItems, str, object]


class FunctionResultContentParam(TypedDict, total=False):
    """A function tool result content block."""

    call_id: Required[str]
    """ID to match the ID from the function call block."""

    result: Required[Result]
    """The result of the tool call."""

    type: Required[Literal["function_result"]]

    is_error: bool
    """Whether the tool call resulted in an error."""

    name: str
    """The name of the tool that was called."""



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ... import _api_client
from ... import _transformers as t
from ... import errors
from ... import types
from .. import pytest_helper


test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name='test_disable_thinking',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash',
            contents=t.t_contents('Explain the monty hall problem.'),
            config={
                'thinking_config': {
                    'thinking_budget': 0},
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_thinking_level',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash',
            contents=t.t_contents('Explain the monty hall problem.'),
            config={
                'thinking_config': {
                    'thinking_level': "LOW"
                },
            },
        ),
    ),
]


pytestmark = pytest_helper.setup(
    file=__file__,
    globals_for_file=globals(),
    test_method='models.generate_content',
    test_table=test_table,
)


def test_thinking_budget(client):
    response = client.models.generate_content(
        model='gemini-2.5-pro',
        contents='What is the sum of natural numbers from 1 to 100?',
        config={
            'thinking_config': {
                'include_thoughts': True,
                'thinking_budget': 10000,
            },
        },
    )
    has_thought = False
    if response.candidates:
        for candidate in response.candidates:
            for part in candidate.content.parts:
                if part.thought:
                    has_thought = True
                    break
    assert has_thought



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import base64
from datetime import datetime
from typing import Any, Optional

import pytest

from ... import _common

# 64 distinct chars in url safe base64.
_URL_SAFE_BASE64 = (
    '-_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
)
_RAW_BYTES = (
    b'\xfb\xf6\x9bq\xd7\x9f\x82\x18\xa3\x92Y\xa7\xa2\x9a\xab\xb2\xdb\xaf\xc3\x1c\xb3\x00\x10\x83\x10Q\x87'
    b' \x92\x8b0\xd3\x8fA\x14\x93QU\x97a\x9d5\xdb~9\xeb\xbf='
)
# 64 distinct chars in normal base64.
_BASE64 = '+/abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
assert base64.b64decode(_BASE64) == _RAW_BYTES


class MyBytesModel(_common.BaseModel):
  image_data: Optional[bytes]


# This test shows that if user pass in url safe base64 string to a bytes field,
# then SDK will return raw bytes in pydantic, model_dump() and __str__()
# and base64 string in to_json_dict() and model_dump(mode='json').
def test_urlsafe_base64_input_success():
  my_data = MyBytesModel(image_data=_URL_SAFE_BASE64)

  # Check output is base64 string.
  assert my_data.to_json_dict() == {'image_data': _URL_SAFE_BASE64}
  assert my_data.model_dump(mode='json') == {'image_data': _URL_SAFE_BASE64}
  assert isinstance(my_data.to_json_dict()['image_data'], str)

  # Check output is bytes.
  assert my_data.model_dump() == {'image_data': _RAW_BYTES}
  assert my_data.image_data == _RAW_BYTES
  assert isinstance(my_data.image_data, bytes)

  # Check print(my_data) will output the bytes string, not the base64 string.
  assert str(my_data) == 'image_data=' + str(_RAW_BYTES)


# This test shows that if user pass in raw bytes to a bytes field,
# then SDK will return raw bytes in pydantic, model_dump() and __str__()
# and base64 string in to_json_dict() and model_dump(mode='json').
def test_raw_bytes_input_success():
  my_data = MyBytesModel(image_data=_RAW_BYTES)

  # Check output is base64 string.
  assert my_data.to_json_dict() == {'image_data': _URL_SAFE_BASE64}
  assert my_data.model_dump(mode='json') == {'image_data': _URL_SAFE_BASE64}
  assert isinstance(my_data.to_json_dict()['image_data'], str)

  # Check output is bytes.
  assert my_data.model_dump() == {'image_data': _RAW_BYTES}
  assert my_data.image_data == _RAW_BYTES
  assert isinstance(my_data.image_data, bytes)

  # Check print(my_data) will output the bytes string, not the base64 string.
  assert str(my_data) == 'image_data=' + str(_RAW_BYTES)


# This test shows that if user pass in invalid base64 string(normal base64 but
# not url safe) to a bytes field, then SDK will raise ValueError.
def test_invalid_base64_pydantic_input_failure():
  with pytest.raises(ValueError, match='Data should be valid base64'):
    MyBytesModel(image_data=_BASE64+'=')  # invalid base64


# This test shows that if user pass in invalid base64 string(normal base64 but
# not url safe) to a bytes field, then SDK will raise ValueError.
def test_invalid_base64_dict_input_failure():
  with pytest.raises(ValueError, match='Data should be valid base64'):
    MyBytesModel.model_validate({'image_data': _BASE64+'='})  # invalid base64


# This test shows that if user pass in normal base64 string(
# not url safe) to a bytes field, then Pydantic can correctly parse the normal
# base64 string to bytes.
def test_normal_base64_pydantic_input_success():
  data = MyBytesModel(image_data=_BASE64)
  assert data.image_data == _RAW_BYTES


# This test shows that if user pass in normal base64 string via(
# not url safe) via dict input with model_validate(), then Pydantic can
# correctly parse the normal base64 string to bytes.
def test_normal_base64_dict_input_success():
  data = MyBytesModel.model_validate({'image_data': _BASE64})
  assert data.image_data == _RAW_BYTES


class MyAnyModel(_common.BaseModel):
  data: Optional[Any]


# This test shows that if user pass in raw bytes to an Any type field,
# then SDK will return raw bytes in pydantic, model_dump() and __str__()
# and base64 string in to_json_dict() and model_dump(mode='json').
def test_any_type_urlsafe_base64_input():
  my_data = MyAnyModel(data=_RAW_BYTES)

  # Check output is base64 string.
  assert my_data.to_json_dict() == {'data': _URL_SAFE_BASE64}
  assert my_data.model_dump(mode='json') == {'data': _URL_SAFE_BASE64}
  assert isinstance(my_data.to_json_dict()['data'], str)

  # Check output is bytes.
  assert my_data.model_dump() == {'data': _RAW_BYTES}
  assert my_data.data == _RAW_BYTES
  assert isinstance(my_data.data, bytes)

  # Check print(my_data) will output the bytes string, not the base64 string.
  assert str(my_data) == 'data=' + str(_RAW_BYTES)


# This test shows if user pass in url safe base64 string to an Any type field,
# then SDK will always return the base64 string.
def test_any_type_urlsafe_base64_input():
  my_data = MyAnyModel(data=_URL_SAFE_BASE64)

  # Check output is base64 string.
  assert my_data.to_json_dict() == {'data': _URL_SAFE_BASE64}
  assert my_data.model_dump(mode='json') == {'data': _URL_SAFE_BASE64}
  assert isinstance(my_data.to_json_dict()['data'], str)

  # Check output is bytes.
  assert my_data.model_dump() == {'data': _URL_SAFE_BASE64}
  assert my_data.data == _URL_SAFE_BASE64
  assert isinstance(my_data.data, str)

  # Check print(my_data) will output the bytes string, not the base64 string.
  assert str(my_data) == "data='" + str(_URL_SAFE_BASE64) + "'"



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Union, Optional
from typing_extensions import Literal, Annotated, TypeAlias

from pydantic import Field as FieldInfo

from .._utils import PropertyInfo
from .._models import BaseModel
from .function import Function
from .allowed_tools import AllowedTools

__all__ = ["Tool", "GoogleSearch", "CodeExecution", "URLContext", "ComputerUse", "MCPServer", "FileSearch"]


class GoogleSearch(BaseModel):
    """A tool that can be used by the model to search Google."""

    type: Literal["google_search"]

    search_types: Optional[List[Literal["web_search", "image_search"]]] = None
    """The types of search grounding to enable."""


class CodeExecution(BaseModel):
    """A tool that can be used by the model to execute code."""

    type: Literal["code_execution"]


class URLContext(BaseModel):
    """A tool that can be used by the model to fetch URL context."""

    type: Literal["url_context"]


class ComputerUse(BaseModel):
    """A tool that can be used by the model to interact with the computer."""

    type: Literal["computer_use"]

    environment: Optional[Literal["browser"]] = None
    """The environment being operated."""

    excluded_predefined_functions: Optional[List[str]] = FieldInfo(alias="excludedPredefinedFunctions", default=None)
    """The list of predefined functions that are excluded from the model call."""


class MCPServer(BaseModel):
    """A MCPServer is a server that can be called by the model to perform actions."""

    type: Literal["mcp_server"]

    allowed_tools: Optional[List[AllowedTools]] = None
    """The allowed tools."""

    headers: Optional[Dict[str, str]] = None
    """Optional: Fields for authentication headers, timeouts, etc., if needed."""

    name: Optional[str] = None
    """The name of the MCPServer."""

    url: Optional[str] = None
    """
    The full URL for the MCPServer endpoint.
    Example: "https://api.example.com/mcp"
    """


class FileSearch(BaseModel):
    """A tool that can be used by the model to search files."""

    type: Literal["file_search"]

    file_search_store_names: Optional[List[str]] = None
    """The file search store names to search."""

    metadata_filter: Optional[str] = None
    """Metadata filter to apply to the semantic retrieval documents and chunks."""

    top_k: Optional[int] = None
    """The number of semantic retrieval chunks to retrieve."""


Tool: TypeAlias = Annotated[
    Union[Function, GoogleSearch, CodeExecution, URLContext, ComputerUse, MCPServer, FileSearch],
    PropertyInfo(discriminator="type"),
]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dataclasses
import functools
import hashlib
import os
import tempfile
from typing import Optional, cast
import uuid

import requests  # type: ignore
import sentencepiece as spm
from sentencepiece import sentencepiece_model_pb2


# Source of truth: https://cloud.google.com/vertex-ai/generative-ai/docs/learn/models
_GEMINI_MODELS_TO_TOKENIZER_NAMES = {
    "gemini-2.5-pro": "gemma3",
    "gemini-2.5-flash": "gemma3",
    "gemini-2.5-flash-lite": "gemma3",
    "gemini-2.0-flash": "gemma3",
    "gemini-2.0-flash-lite": "gemma3",
}
_GEMINI_STABLE_MODELS_TO_TOKENIZER_NAMES = {
    "gemini-2.5-pro-preview-06-05": "gemma3",
    "gemini-2.5-pro-preview-05-06": "gemma3",
    "gemini-2.5-pro-exp-03-25": "gemma3",
    "gemini-live-2.5-flash": "gemma3",
    "gemini-2.5-flash-preview-05-20": "gemma3",
    "gemini-2.5-flash-preview-04-17": "gemma3",
    "gemini-2.5-flash-lite-preview-06-17": "gemma3",
    "gemini-2.0-flash-001": "gemma3",
    "gemini-2.0-flash-lite-001": "gemma3",
    "gemini-3-pro-preview": "gemma3",
}


@dataclasses.dataclass(frozen=True)
class _TokenizerConfig:
  model_url: str
  model_hash: str


# TODO: update gemma3 tokenizer
_TOKENIZERS = {
    "gemma2": _TokenizerConfig(
        model_url="https://raw.githubusercontent.com/google/gemma_pytorch/33b652c465537c6158f9a472ea5700e5e770ad3f/tokenizer/tokenizer.model",
        model_hash=(
            "61a7b147390c64585d6c3543dd6fc636906c9af3865a5548f27f31aee1d4c8e2"
        ),
    ),
    "gemma3": _TokenizerConfig(
        model_url="https://raw.githubusercontent.com/google/gemma_pytorch/014acb7ac4563a5f77c76d7ff98f31b568c16508/tokenizer/gemma3_cleaned_262144_v2.spiece.model",
        model_hash=(
            "1299c11d7cf632ef3b4e11937501358ada021bbdf7c47638d13c0ee982f2e79c"
        ),
    ),
}


def _load_file(file_url_path: str) -> bytes:
  """Loads file bytes from the given file url path."""
  resp = requests.get(file_url_path)
  resp.raise_for_status()
  return cast(bytes, resp.content)


def _is_valid_model(*, model_data: bytes, expected_hash: str) -> bool:
  """Returns true if the content is valid by checking the hash."""
  if not expected_hash:
    raise ValueError("expected_hash is required")
  return hashlib.sha256(model_data).hexdigest() == expected_hash


def _maybe_remove_file(file_path: str) -> None:
  """Removes the file if exists."""
  if not os.path.exists(file_path):
    return
  try:
    os.remove(file_path)
  except OSError:
    # Don't raise if we cannot remove file.
    pass


def _maybe_load_from_cache(
    *, file_path: str, expected_hash: str
) -> Optional[bytes]:
  """Loads the content from the cache path."""
  if not os.path.exists(file_path):
    return None
  with open(file_path, "rb") as f:
    content = f.read()
  if _is_valid_model(model_data=content, expected_hash=expected_hash):
    return content

  # Cached file corrupted.
  _maybe_remove_file(file_path)
  return None


def _maybe_save_to_cache(
    *, cache_dir: str, cache_path: str, content: bytes
) -> None:
  """Saves the content to the cache path."""
  try:
    os.makedirs(cache_dir, exist_ok=True)
    tmp_path = cache_dir + "." + str(uuid.uuid4()) + ".tmp"
    with open(tmp_path, "wb") as f:
      f.write(content)
    os.rename(tmp_path, cache_path)
  except OSError:
    # Don't raise if we cannot write file.
    pass


def _load_from_url(*, file_url: str, expected_hash: str) -> bytes:
  """Loads model bytes from the given file url."""
  content = _load_file(file_url)
  if not _is_valid_model(model_data=content, expected_hash=expected_hash):
    actual_hash = hashlib.sha256(content).hexdigest()
    raise ValueError(
        "Downloaded model file is corrupted."
        f" Expected hash {expected_hash}. Got file hash {actual_hash}."
    )
  return content


def _load(*, file_url: str, expected_hash: str) -> bytes:
  """Loads model bytes from the given file url.

  1. If the find local cached file for the given url and the cached file hash
     matches the expected hash, the cached file is returned.
  2. If local cached file is not found or the hash does not match, the file is
     downloaded from the given url. And write to local cache and return the
     file bytes.
  3. If the file downloaded from the given url does not match the expected
     hash, raise ValueError.

  Args:
      file_url: The url of the file to load.
      expected_hash: The expected hash of the file.

  Returns:
      The file bytes.
  """
  model_dir = os.path.join(tempfile.gettempdir(), "vertexai_tokenizer_model")
  filename = hashlib.sha1(file_url.encode()).hexdigest()
  model_path = os.path.join(model_dir, filename)

  model_data = _maybe_load_from_cache(
      file_path=model_path, expected_hash=expected_hash
  )
  if not model_data:
    model_data = _load_from_url(file_url=file_url, expected_hash=expected_hash)

  _maybe_save_to_cache(
      cache_dir=model_dir, cache_path=model_path, content=model_data
  )
  return model_data


def _load_model_proto_bytes(tokenizer_name: str) -> bytes:
  """Loads model proto bytes from the given tokenizer name."""
  if tokenizer_name not in _TOKENIZERS:
    raise ValueError(
        f"Tokenizer {tokenizer_name} is not supported."
        f"Supported tokenizers: {list(_TOKENIZERS.keys())}"
    )
  return _load(
      file_url=_TOKENIZERS[tokenizer_name].model_url,
      expected_hash=_TOKENIZERS[tokenizer_name].model_hash,
  )


@functools.lru_cache()
def load_model_proto(
    tokenizer_name: str,
) -> sentencepiece_model_pb2.ModelProto:
  """Loads model proto from the given tokenizer name."""
  model_proto = sentencepiece_model_pb2.ModelProto()
  model_proto.ParseFromString(_load_model_proto_bytes(tokenizer_name))
  return model_proto


def get_tokenizer_name(model_name: str) -> str:
  """Gets the tokenizer name for the given model name."""
  if model_name in _GEMINI_MODELS_TO_TOKENIZER_NAMES.keys():
    return _GEMINI_MODELS_TO_TOKENIZER_NAMES[model_name]
  if model_name in _GEMINI_STABLE_MODELS_TO_TOKENIZER_NAMES.keys():
    return _GEMINI_STABLE_MODELS_TO_TOKENIZER_NAMES[model_name]
  raise ValueError(
      f"Model {model_name} is not supported. Supported models: {', '.join(_GEMINI_MODELS_TO_TOKENIZER_NAMES.keys())}, {', '.join(_GEMINI_STABLE_MODELS_TO_TOKENIZER_NAMES.keys())}.\n"  # pylint: disable=line-too-long
  )


@functools.lru_cache()
def get_sentencepiece(tokenizer_name: str) -> spm.SentencePieceProcessor:
  """Loads sentencepiece tokenizer from the given tokenizer name."""
  processor = spm.SentencePieceProcessor()
  processor.LoadFromSerializedProto(_load_model_proto_bytes(tokenizer_name))
  return processor



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import time
from unittest.mock import MagicMock, patch
import pytest
from .api_client import BaseApiClient


@patch('genai.api_client.BaseApiClient._build_request')
@patch('genai.api_client.BaseApiClient._request')
def test_request_streamed_non_blocking(mock_request, mock_build_request):
  api_client = BaseApiClient(api_key='test_api_key')
  http_method = 'GET'
  path = 'test/path'
  request_dict = {'key': 'value'}

  mock_http_request = MagicMock()
  mock_build_request.return_value = mock_http_request

  def delayed_segments():
    chunks = ['{"chunk": 1}', '{"chunk": 2}', '{"chunk": 3}']
    for chunk in chunks:
      time.sleep(0.1)  # 100ms delay
      yield chunk

  mock_response = MagicMock()
  mock_response.segments.side_effect = delayed_segments
  mock_request.return_value = mock_response

  chunks = []
  start_time = time.time()
  for chunk in api_client.request_streamed(http_method, path, request_dict):
    chunks.append(chunk)
    assert len(chunks) <= 3
  end_time = time.time()

  mock_build_request.assert_called_once_with(
      http_method, path, request_dict, None
  )
  mock_request.assert_called_once_with(mock_http_request, stream=True)
  assert chunks == ['{"chunk": 1}', '{"chunk": 2}', '{"chunk": 3}']
  assert end_time - start_time > 0.3


@patch('genai.api_client.BaseApiClient._build_request')
@patch('genai.api_client.BaseApiClient._async_request')
@pytest.mark.asyncio
async def test_async_request(mock_async_request, mock_build_request):
  api_client = ApiClient(api_key='test_api_key')
  http_method = 'GET'
  path = 'test/path'
  request_dict = {'key': 'value'}

  mock_http_request = MagicMock()
  mock_build_request.return_value = mock_http_request

  class MockResponse:

    def __init__(self, text):
      self.text = text

  async def delayed_response(http_request, stream):
    await asyncio.sleep(0.1)  # 100ms delay
    return MockResponse('value')

  mock_async_request.side_effect = delayed_response

  async_coroutine1 = api_client.async_request(http_method, path, request_dict)
  async_coroutine2 = api_client.async_request(http_method, path, request_dict)
  async_coroutine3 = api_client.async_request(http_method, path, request_dict)

  start_time = time.time()
  results = await asyncio.gather(
      async_coroutine1, async_coroutine2, async_coroutine3
  )
  end_time = time.time()

  mock_build_request.assert_called_with(http_method, path, request_dict, None)
  assert mock_build_request.call_count == 3
  mock_async_request.assert_called_with(
      http_request=mock_http_request, stream=False
  )
  assert mock_async_request.call_count == 3
  assert results == ['value', 'value', 'value']
  assert 0.1 <= end_time - start_time < 0.15


@patch('genai.api_client.BaseApiClient._build_request')
@patch('genai.api_client.BaseApiClient._async_request')
@pytest.mark.asyncio
async def test_async_request_streamed_non_blocking(
    mock_async_request, mock_build_request
):
  api_client = ApiClient(api_key='test_api_key')
  http_method = 'GET'
  path = 'test/path'
  request_dict = {'key': 'value'}

  mock_http_request = MagicMock()
  mock_build_request.return_value = mock_http_request

  class MockResponse:

    def __init__(self, segments):
      self._segments = segments

    # should mock async generator here but source code combines sync and async streaming in one segment method.
    # TODO: fix the above
    def segments(self):
      for segment in self._segments:
        time.sleep(0.1)  # 100ms delay
        yield segment

  async def delayed_response(http_request, stream):
    return MockResponse(['{"chunk": 1}', '{"chunk": 2}', '{"chunk": 3}'])

  mock_async_request.side_effect = delayed_response

  chunks = []
  start_time = time.time()
  async for chunk in await api_client.async_request_streamed(
      http_method, path, request_dict
  ):
    chunks.append(chunk)
    assert len(chunks) <= 3
  end_time = time.time()

  mock_build_request.assert_called_once_with(
      http_method, path, request_dict, None
  )
  mock_async_request.assert_called_once_with(
      http_request=mock_http_request, stream=True
  )
  assert chunks == ['{"chunk": 1}', '{"chunk": 2}', '{"chunk": 3}']
  assert end_time - start_time > 0.3



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for live.py."""

import contextlib
import json
from unittest import mock

import pytest
from websockets import client

from .. import pytest_helper
from ... import client as gl_client
from ... import live
from ... import types


def exception_if_mldev(vertexai, exception_type: type[Exception]):
  if vertexai:
    return contextlib.nullcontext()
  else:
    return pytest.raises(exception_type)


def mock_api_client(vertexai=False):
  api_client = mock.MagicMock(spec=gl_client.BaseApiClient)
  api_client.api_key = 'TEST_API_KEY'
  api_client._host = lambda: 'test_host'
  api_client._http_options = {'headers': {}}  # Ensure headers exist
  api_client.vertexai = vertexai
  api_client._api_client = api_client
  return api_client


@pytest.fixture
def mock_websocket():
  websocket = mock.AsyncMock(spec=client.ClientConnection)
  websocket.send = mock.AsyncMock()
  websocket.recv = mock.AsyncMock(
      return_value='{"serverContent": {"turnComplete": true}}'
  )  # Default response
  websocket.close = mock.AsyncMock()
  return websocket


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_function_response_dict(mock_websocket, vertexai):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(
      api_client=api_client, websocket=mock_websocket
  )

  input = {
      'name': 'get_current_weather',
      'response': {'temperature': 14.5, 'unit': 'C'},
  }

  if not vertexai:
    input['id'] = 'some-id'

  await session.send_tool_response(function_responses=input)

  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'tool_response' in sent_data

  assert (
      sent_data['tool_response']['functionResponses'][0]['name']
      == 'get_current_weather'
  )
  assert (
      sent_data['tool_response']['functionResponses'][0]['response'][
          'temperature'
      ]
      == 14.5
  )
  assert (
      sent_data['tool_response']['functionResponses'][0]['response']['unit']
      == 'C'
  )


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_function_response(mock_websocket, vertexai):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )

  input = types.FunctionResponse(
      name='get_current_weather',
      response={
          'temperature': 14.5,
          'unit': 'C',
          'user_name': 'test_user_name',
          'userEmail': 'test_user_email',
      },
  )
  if not vertexai:
    input.id = 'some-id'

  await session.send_tool_response(function_responses=input)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'tool_response' in sent_data

  assert (
      sent_data['tool_response']['functionResponses'][0]['name']
      == 'get_current_weather'
  )
  assert (
      sent_data['tool_response']['functionResponses'][0]['response']
      == input.response
  )


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_function_response_scheduling(mock_websocket, vertexai):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(api_client=api_client, websocket=mock_websocket)

  input = types.FunctionResponse(
      name='get_current_weather',
      response={'temperature': 14.5, 'unit': 'C'},
      will_continue=True,
      scheduling=types.FunctionResponseScheduling.SILENT,
  )
  if not vertexai:
    input.id = 'some-id'

  await session.send_tool_response(function_responses=input)

  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'tool_response' in sent_data

  assert pytest_helper.get_value_ignore_key_case(
      sent_data['tool_response']['functionResponses'][0], 'will_continue'
  )
  assert (
      sent_data['tool_response']['functionResponses'][0]['scheduling']
      == 'SILENT'
  )


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_function_response_list(mock_websocket, vertexai):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )

  input1 = {
      'name': 'get_current_weather',
      'response': {'temperature': 14.5, 'unit': 'C'},
  }
  input2 = {
      'name': 'get_current_weather',
      'response': {'temperature': 99.9, 'unit': 'C'},
  }

  if not vertexai:
    input1['id'] = '1'
    input2['id'] = '2'

  await session.send_tool_response(function_responses=[input1, input2])
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'tool_response' in sent_data

  assert len(sent_data['tool_response']['functionResponses']) == 2
  assert (
      sent_data['tool_response']['functionResponses'][0]['response'][
          'temperature'
      ]
      == 14.5
  )
  assert (
      sent_data['tool_response']['functionResponses'][1]['response'][
          'temperature'
      ]
      == 99.9
  )


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_missing_id(mock_websocket, vertexai):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(
      api_client=api_client, websocket=mock_websocket
  )

  input1 = {
      'name': 'get_current_weather',
      'response': {'temperature': 14.5, 'unit': 'C'},
      'id': '1',
  }
  input2 = {
      'name': 'get_current_weather',
      'response': {'temperature': 99.9, 'unit': 'C'},
  }

  if not vertexai:
    with pytest.raises(ValueError, match=".*must have.*"):
      await session.send_tool_response(function_responses=[input1, input2])



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import typing
from typing import Any

from ... import _mcp_utils
from ... import types

if typing.TYPE_CHECKING:
  from mcp.types import Tool as McpTool
  from mcp import ClientSession as McpClientSession
else:
  McpTool: typing.Type = Any
  McpClientSession: typing.Type = Any
  try:
    from mcp.types import Tool as McpTool
    from mcp import ClientSession as McpClientSession
  except ImportError:
    McpTool = None
    McpClientSession = None


def test_mcp_tools():
  """Test whether the list of tools contains any MCP tools."""
  if McpTool is None:
    return
  mcp_tools = [
      McpTool(
          name='tool',
          description='tool-description',
          inputSchema={
              'type': 'OBJECT',
              'properties': {
                  'key1': {'type': 'STRING'},
                  'key2': {'type': 'NUMBER'},
              },
          },
      ),
  ]
  assert _mcp_utils.has_mcp_tool_usage(mcp_tools)


def test_mcp_client_session():
  """Test whether the list of tools contains any MCP tools."""

  class MockMcpClientSession(McpClientSession):

    def __init__(self):
      self._read_stream = None
      self._write_stream = None

  mcp_tools = [
      MockMcpClientSession(),
  ]
  assert _mcp_utils.has_mcp_tool_usage(mcp_tools)


def test_no_mcp_tools():
  if McpClientSession is None:
    return
  """Test whether the list of tools contains any MCP tools."""
  gemini_tools = [
      types.Tool(
          function_declarations=[
              types.FunctionDeclaration(
                  name='tool',
                  description='tool-description',
                  parameters=types.Schema(
                      type='OBJECT',
                      properties={},
                  ),
              ),
          ],
      ),
  ]
  assert not _mcp_utils.has_mcp_tool_usage(gemini_tools)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for get_function_response_parts."""

import typing
from typing import Any
import pytest
from ..._extra_utils import get_function_response_parts, get_function_response_parts_async
from ...errors import UnsupportedFunctionError
from ...types import Candidate
from ...types import Content
from ...types import FunctionCall
from ...types import FunctionResponse
from ...types import GenerateContentResponse
from ...types import Part

_is_mcp_imported = False
if typing.TYPE_CHECKING:
  from mcp import types as mcp_types
  from mcp import ClientSession as McpClientSession
  from ..._adapters import McpToGenAiToolAdapter

  _is_mcp_imported = True
else:
  McpClientSession: typing.Type = Any
  McpToGenAiToolAdapter: typing.Type = Any
  try:
    from mcp import types as mcp_types
    from mcp import ClientSession as McpClientSession
    from ..._adapters import McpToGenAiToolAdapter

    _is_mcp_imported = True
  except ImportError:
    McpClientSession = None
    McpToGenAiToolAdapter = None


def test_integer_value():
  def func_under_test(a: int) -> int:
    return a + 1

  response = GenerateContentResponse(
      candidates=[
          Candidate(
              content=Content(
                  parts=[
                      Part(
                          function_call=FunctionCall(
                              name='func_under_test',
                              args={'a': 1},
                          )
                      )
                  ]
              )
          )
      ]
  )
  function_map = {'func_under_test': func_under_test}
  expected_parts = [
      Part(
          function_response=FunctionResponse(
              name='func_under_test',
              response={'result': 2},
          )
      )
  ]
  actual_parts = get_function_response_parts(response, function_map)

  for actual_part, expected_part in zip(actual_parts, expected_parts):
    assert actual_part.model_dump_json(
        exclude_none=True
    ) == expected_part.model_dump_json(exclude_none=True)


def test_float_value():
  def func_under_test(a: float) -> float:
    return a + 1.0

  response = GenerateContentResponse(
      candidates=[
          Candidate(
              content=Content(
                  parts=[
                      Part(
                          function_call=FunctionCall(
                              name='func_under_test',
                              args={'a': 1.0},
                          )
                      )
                  ]
              )
          )
      ]
  )
  function_map = {'func_under_test': func_under_test}
  expected_parts = [
      Part(
          function_response=FunctionResponse(
              name='func_under_test',
              response={'result': 2.0},
          )
      )
  ]
  actual_parts = get_function_response_parts(response, function_map)

  for actual_part, expected_part in zip(actual_parts, expected_parts):
    assert actual_part.model_dump_json(
        exclude_none=True
    ) == expected_part.model_dump_json(exclude_none=True)


def test_string_value():
  def func_under_test(a: str) -> str:
    return a + '1'

  response = GenerateContentResponse(
      candidates=[
          Candidate(
              content=Content(
                  parts=[
                      Part(
                          function_call=FunctionCall(
                              name='func_under_test',
                              args={'a': '1.0'},
                          )
                      )
                  ]
              )
          )
      ]
  )
  function_map = {'func_under_test': func_under_test}
  expected_parts = [
      Part(
          function_response=FunctionResponse(
              name='func_under_test',
              response={'result': '1.01'},
          )
      )
  ]
  actual_parts = get_function_response_parts(response, function_map)

  for actual_part, expected_part in zip(actual_parts, expected_parts):
    assert actual_part.model_dump_json(
        exclude_none=True
    ) == expected_part.model_dump_json(exclude_none=True)


@pytest.mark.asyncio
async def test_mcp_tool():
  if not _is_mcp_imported:
    return

  class MockMcpClientSession(McpClientSession):

    def __init__(self):
      self._read_stream = None
      self._write_stream = None

    async def call_tool(self, name: str, arguments: dict[str, Any]) -> Any:
      return mcp_types.CallToolResult(
          content=[mcp_types.TextContent(type='text', text='1.01')]
      )

  mcp_to_genai_tool_adapter = McpToGenAiToolAdapter(
      session=MockMcpClientSession(),
      list_tools_result=mcp_types.ListToolsResult(tools=[]),
  )
  response = GenerateContentResponse(
      candidates=[
          Candidate(
              content=Content(
                  parts=[
                      Part(
                          function_call=FunctionCall(
                              name='tool',
                              args={'key1': 'value1', 'key2': 1},
                          )
                      )
                  ]
              )
          )
      ]
  )
  function_map = {'tool': mcp_to_genai_tool_adapter}
  expected_parts = [
      Part(
          function_response=FunctionResponse(
              name='tool',
              response={
                  'result': {
                      'content': [{'type': 'text', 'text': '1.01'}],
                      'isError': False,
                  }
              },
          )
      )
  ]
  actual_parts = await get_function_response_parts_async(response, function_map)

  for actual_part, expected_part in zip(actual_parts, expected_parts):
    assert actual_part.model_dump_json(
        exclude_none=True
    ) == expected_part.model_dump_json(exclude_none=True)


@pytest.mark.asyncio
async def test_mcp_tool_error():
  if not _is_mcp_imported:
    return

  class MockMcpClientSession(McpClientSession):

    def __init__(self):
      self._read_stream = None
      self._write_stream = None

    async def call_tool(self, name: str, arguments: dict[str, Any]) -> Any:
      return mcp_types.CallToolResult(
          content=[mcp_types.TextContent(type='text', text='Internal error')],
          isError=True,
      )

  mcp_to_genai_tool_adapter = McpToGenAiToolAdapter(
      session=MockMcpClientSession(),
      list_tools_result=mcp_types.ListToolsResult(tools=[]),
  )
  response = GenerateContentResponse(
      candidates=[
          Candidate(
              content=Content(
                  parts=[
                      Part(
                          function_call=FunctionCall(
                              name='tool',
                              args={'key1': 'value1', 'key2': 1},
                          )
                      )
                  ]
              )
          )
      ]
  )
  function_map = {'tool': mcp_to_genai_tool_adapter}
  expected_parts = [
      Part(
          function_response=FunctionResponse(
              name='tool',
              response={
                  'error': {
                      'content': [{'type': 'text', 'text': 'Internal error'}],
                      'isError': True,
                  }
              },
          )
      )
  ]
  actual_parts = await get_function_response_parts_async(response, function_map)

  for actual_part, expected_part in zip(actual_parts, expected_parts):
    assert actual_part.model_dump_json(
        exclude_none=True
    ) == expected_part.model_dump_json(exclude_none=True)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Mapping
from typing_extensions import Self, override

import httpx

from . import _exceptions
from ._qs import Querystring
from ._types import (
    Omit,
    Headers,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
    not_given,
)
from ._utils import is_given, get_async_library
from ._compat import cached_property
from ._models import FinalRequestOptions
from ._version import __version__
from ._streaming import Stream as Stream, AsyncStream as AsyncStream
from ._exceptions import APIStatusError
from ._base_client import (
    DEFAULT_MAX_RETRIES,
    SyncAPIClient,
    AsyncAPIClient,
)
from ._client_adapter import GeminiNextGenAPIClientAdapter, AsyncGeminiNextGenAPIClientAdapter

if TYPE_CHECKING:
    from .resources import interactions
    from .resources.interactions import InteractionsResource, AsyncInteractionsResource

__all__ = [
    "Timeout",
    "Transport",
    "ProxiesTypes",
    "RequestOptions",
    "GeminiNextGenAPIClient",
    "AsyncGeminiNextGenAPIClient",
    "Client",
    "AsyncClient",
]


class GeminiNextGenAPIClient(SyncAPIClient):
    # client options
    api_key: str | None
    api_version: str
    client_adapter: GeminiNextGenAPIClientAdapter | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        api_version: str | None = "v1beta",
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        client_adapter: GeminiNextGenAPIClientAdapter | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous GeminiNextGenAPIClient client instance.

        This automatically infers the `api_key` argument from the `GEMINI_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("GEMINI_API_KEY")
        self.api_key = api_key

        if api_version is None:
            api_version = "v1beta"
        self.api_version = api_version

        if base_url is None:
            base_url = os.environ.get("GEMINI_NEXT_GEN_API_BASE_URL")
        if base_url is None:
            base_url = f"https://generativelanguage.googleapis.com"

        self.client_adapter = client_adapter

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._default_stream_cls = Stream

    @cached_property
    def interactions(self) -> InteractionsResource:
        from .resources.interactions import InteractionsResource

        return InteractionsResource(self)

    @cached_property
    def with_raw_response(self) -> GeminiNextGenAPIClientWithRawResponse:
        return GeminiNextGenAPIClientWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> GeminiNextGenAPIClientWithStreamedResponse:
        return GeminiNextGenAPIClientWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        if api_key is None:
            return {}
        return {"x-goog-api-key": api_key}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    @override
    def _validate_headers(self, headers: Headers, custom_headers: Headers) -> None:
        if headers.get("Authorization") or custom_headers.get("Authorization") or isinstance(custom_headers.get("Authorization"), Omit):
            return
        if self.api_key and headers.get("x-goog-api-key"):
            return
        if custom_headers.get("x-goog-api-key") or isinstance(custom_headers.get("x-goog-api-key"), Omit):
            return

        raise TypeError(
            '"Could not resolve authentication method. Expected the api_key to be set. Or for the `x-goog-api-key` headers to be explicitly omitted"'
        )
    
    @override
    def _prepare_options(self, options: FinalRequestOptions) -> FinalRequestOptions:
        if not self.client_adapter or not self.client_adapter.is_vertex_ai():
            return options

        headers = options.headers or {}
        has_auth = headers.get("Authorization") or headers.get("x-goog-api-key")   # pytype: disable=attribute-error
        if has_auth:
            return options

        adapted_headers = self.client_adapter.get_auth_headers()
        if adapted_headers:
            options.headers = {
                **adapted_headers,
                **headers
            }
        return options
    
    def copy(
        self,
        *,
        api_key: str | None = None,
        api_version: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        client_adapter: GeminiNextGenAPIClientAdapter | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            api_version=api_version or self.api_version,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            client_adapter=self.client_adapter or client_adapter,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def _get_api_version_path_param(self) -> str:
        return self.api_version

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class AsyncGeminiNextGenAPIClient(AsyncAPIClient):
    # client options
    api_key: str | None
    api_version: str
    client_adapter: AsyncGeminiNextGenAPIClientAdapter | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        api_version: str | None = "v1beta",
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultAsyncHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        client_adapter: AsyncGeminiNextGenAPIClientAdapter | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async AsyncGeminiNextGenAPIClient client instance.

        This automatically infers the `api_key` argument from the `GEMINI_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("GEMINI_API_KEY")
        self.api_key = api_key

        if api_version is None:
            api_version = "v1beta"
        self.api_version = api_version

        if base_url is None:
            base_url = os.environ.get("GEMINI_NEXT_GEN_API_BASE_URL")
        if base_url is None:
            base_url = f"https://generativelanguage.googleapis.com"

        self.client_adapter = client_adapter

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._default_stream_cls = AsyncStream

    @cached_property
    def interactions(self) -> AsyncInteractionsResource:
        from .resources.interactions import AsyncInteractionsResource

        return AsyncInteractionsResource(self)

    @cached_property
    def with_raw_response(self) -> AsyncGeminiNextGenAPIClientWithRawResponse:
        return AsyncGeminiNextGenAPIClientWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncGeminiNextGenAPIClientWithStreamedResponse:
        return AsyncGeminiNextGenAPIClientWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        if api_key is None:
            return {}
        return {"x-goog-api-key": api_key}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": f"async:{get_async_library()}",
            **self._custom_headers,
        }

    @override
    def _validate_headers(self, headers: Headers, custom_headers: Headers) -> None:
        if headers.get("Authorization") or custom_headers.get("Authorization") or isinstance(custom_headers.get("Authorization"), Omit):
            return
        if self.api_key and headers.get("x-goog-api-key"):
            return
        if custom_headers.get("x-goog-api-key") or isinstance(custom_headers.get("x-goog-api-key"), Omit):
            return

        raise TypeError(
            '"Could not resolve authentication method. Expected the api_key to be set. Or for the `x-goog-api-key` headers to be explicitly omitted"'
        )
    
    @override
    async def _prepare_options(self, options: FinalRequestOptions) -> FinalRequestOptions:
        if not self.client_adapter or not self.client_adapter.is_vertex_ai():
            return options

        headers = options.headers or {}
        has_auth = headers.get("Authorization") or headers.get("x-goog-api-key")     # pytype: disable=attribute-error
        if has_auth:
            return options

        adapted_headers = await self.client_adapter.async_get_auth_headers()
        if adapted_headers:
            options.headers = {
                **adapted_headers,
                **headers
            }
        return options

    def copy(
        self,
        *,
        api_key: str | None = None,
        api_version: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        client_adapter: AsyncGeminiNextGenAPIClientAdapter | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            api_version=api_version or self.api_version,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            client_adapter=self.client_adapter or client_adapter,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def _get_api_version_path_param(self) -> str:
        return self.api_version

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class GeminiNextGenAPIClientWithRawResponse:
    _client: GeminiNextGenAPIClient

    def __init__(self, client: GeminiNextGenAPIClient) -> None:
        self._client = client

    @cached_property
    def interactions(self) -> interactions.InteractionsResourceWithRawResponse:
        from .resources.interactions import InteractionsResourceWithRawResponse

        return InteractionsResourceWithRawResponse(self._client.interactions)


class AsyncGeminiNextGenAPIClientWithRawResponse:
    _client: AsyncGeminiNextGenAPIClient

    def __init__(self, client: AsyncGeminiNextGenAPIClient) -> None:
        self._client = client

    @cached_property
    def interactions(self) -> interactions.AsyncInteractionsResourceWithRawResponse:
        from .resources.interactions import AsyncInteractionsResourceWithRawResponse

        return AsyncInteractionsResourceWithRawResponse(self._client.interactions)


class GeminiNextGenAPIClientWithStreamedResponse:
    _client: GeminiNextGenAPIClient

    def __init__(self, client: GeminiNextGenAPIClient) -> None:
        self._client = client

    @cached_property
    def interactions(self) -> interactions.InteractionsResourceWithStreamingResponse:
        from .resources.interactions import InteractionsResourceWithStreamingResponse

        return InteractionsResourceWithStreamingResponse(self._client.interactions)


class AsyncGeminiNextGenAPIClientWithStreamedResponse:
    _client: AsyncGeminiNextGenAPIClient

    def __init__(self, client: AsyncGeminiNextGenAPIClient) -> None:
        self._client = client

    @cached_property
    def interactions(self) -> interactions.AsyncInteractionsResourceWithStreamingResponse:
        from .resources.interactions import AsyncInteractionsResourceWithStreamingResponse

        return AsyncInteractionsResourceWithStreamingResponse(self._client.interactions)


Client = GeminiNextGenAPIClient

AsyncClient = AsyncGeminiNextGenAPIClient



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from typing import Any
from datetime import datetime
from typing_extensions import override

import pydantic

from .._compat import model_dump


def openapi_dumps(obj: Any) -> bytes:
    """
    Serialize an object to UTF-8 encoded JSON bytes.

    Extends the standard json.dumps with support for additional types
    commonly used in the SDK, such as `datetime`, `pydantic.BaseModel`, etc.
    """
    return json.dumps(
        obj,
        cls=_CustomEncoder,
        # Uses the same defaults as httpx's JSON serialization
        ensure_ascii=False,
        separators=(",", ":"),
        allow_nan=False,
    ).encode()


class _CustomEncoder(json.JSONEncoder):
    @override
    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, pydantic.BaseModel):
            return model_dump(o, exclude_unset=True, mode="json", by_alias=True)
        return super().default(o)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from typing import Optional

from .types import HttpOptions

_default_base_gemini_url = None
_default_base_vertex_url = None


def set_default_base_urls(
    gemini_url: Optional[str], vertex_url: Optional[str]
) -> None:
  """Overrides the base URLs for the Gemini API and Vertex AI API."""
  global _default_base_gemini_url, _default_base_vertex_url
  _default_base_gemini_url = gemini_url
  _default_base_vertex_url = vertex_url


def get_base_url(
    vertexai: bool,
    http_options: Optional[HttpOptions] = None,
) -> Optional[str]:
  """Returns the default base URL based on the following priority.

  1. Base URLs set via HttpOptions.
  2. Base URLs set via the latest call to setDefaultBaseUrls.
  3. Base URLs set via environment variables.
  """
  if http_options and http_options.base_url:
    return http_options.base_url

  if vertexai:
    return _default_base_vertex_url or os.getenv('GOOGLE_VERTEX_BASE_URL')
  else:
    return _default_base_gemini_url or os.getenv('GOOGLE_GEMINI_BASE_URL')



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for convert_number_values_for_function_call_args."""

from ..._extra_utils import convert_number_values_for_function_call_args


def test_integer_value():
  assert convert_number_values_for_function_call_args(1) == 1


def test_float_value():
  assert convert_number_values_for_function_call_args(1.0) == 1


def test_string_value():
  assert convert_number_values_for_function_call_args('1.0') == '1.0'


def test_boolean_value():
  assert convert_number_values_for_function_call_args(True) is True


def test_none_value():
  assert convert_number_values_for_function_call_args(None) is None


def test_float_value_with_decimal():
  assert convert_number_values_for_function_call_args(1.1) == 1.1


def test_dict_value():
  assert convert_number_values_for_function_call_args(
      {'key1': 1.0, 'key2': 1.1}
  ) == {'key1': 1, 'key2': 1.1}


def test_list_value():
  assert convert_number_values_for_function_call_args([1.0, 1.1, 1.2]) == [
      1,
      1.1,
      1.2,
  ]


def test_nested_value():
  assert convert_number_values_for_function_call_args(
      {'key1': 1.0, 'key2': {'key3': 1.0, 'key4': [1.2, 2.0]}}
  ) == {'key1': 1, 'key2': {'key3': 1, 'key4': [1.2, 2]}}



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests schema processing methods in the _transformers module."""

import copy
from typing import Union

import pydantic
import pytest

from ... import _transformers
from ... import client as google_genai_client_module
from ... import types


class CurrencyInfo(pydantic.BaseModel):
  name: str
  code: str
  symbol: str


currency_info_fields = CurrencyInfo.model_fields


class CountryInfo(pydantic.BaseModel):
  name: str
  population: int
  capital: str
  continent: str
  gdp: int
  official_language: str
  total_area_sq_mi: int


country_info_fields = CountryInfo.model_fields


class CountryInfoWithCurrency(pydantic.BaseModel):
  name: str
  population: int
  capital: str
  continent: str
  gdp: int
  official_language: str
  total_area_sq_mi: int
  currency: CurrencyInfo


nested_country_info_fields = CountryInfoWithCurrency.model_fields


class CountryInfoWithNullFields(pydantic.BaseModel):
  name: str
  population: Union[int, None] = None


class CountryInfoWithDefaultValue(pydantic.BaseModel):
  name: str
  population: int = 0


class CountryInfoWithAnyOf(pydantic.BaseModel):
  name: str
  restaurants_per_capita: Union[int, float]


@pytest.fixture
def client(use_vertex):
  if use_vertex:
    yield google_genai_client_module.Client(
        vertexai=use_vertex, project='test-project', location='test-location'
    )
  else:
    yield google_genai_client_module.Client(
        vertexai=use_vertex, api_key='test-api-key'
    )


@pytest.mark.parametrize('use_vertex', [True, False])
def test_build_schema_for_list_of_pydantic_schema(client):
  """Tests _build_schema() when list[pydantic.BaseModel] is provided to response_schema."""

  list_schema = _transformers.t_schema(client, CountryInfo).model_dump()

  assert isinstance(list_schema, dict)

  for field_name in list_schema['properties']:
    assert 'title' in list_schema['properties'][field_name]
    assert 'type' in list_schema['properties'][field_name]
    assert field_name in country_info_fields
    field_type_str = country_info_fields[field_name].annotation.__name__
    assert (
        list_schema['properties'][field_name]['type']
        .lower()
        .startswith(field_type_str.lower())
    )
    assert 'required' in list_schema
    assert list_schema['required'] == list(country_info_fields.keys())


@pytest.mark.parametrize('use_vertex', [True, False])
def test_build_schema_for_list_of_nested_pydantic_schema(client):
  """Tests _build_schema() when list[pydantic.BaseModel] is provided to response_schema and the pydantic.BaseModel has nested pydantic fields."""
  list_schema = _transformers.t_schema(
      client, CountryInfoWithCurrency
  ).model_dump()

  assert isinstance(list_schema, dict)

  for field_name in list_schema['properties']:
    assert 'title' in list_schema['properties'][field_name]
    assert 'type' in list_schema['properties'][field_name]
    assert field_name in nested_country_info_fields

  # Tested nested schema was created
  assert 'properties' in list_schema['properties']['currency']

  for field_name in list_schema['properties']['currency']['properties']:
    assert field_name in currency_info_fields


@pytest.mark.parametrize('use_vertex', [True, False])
def test_t_schema_for_pydantic_schema(client):
  """Tests t_schema when pydantic.BaseModel is passed to response_schema."""
  transformed_schema = _transformers.t_schema(client, CountryInfo)
  assert isinstance(transformed_schema, types.Schema)
  for schema_property in transformed_schema.properties:
    assert schema_property in country_info_fields
    assert isinstance(
        transformed_schema.properties[schema_property], types.Schema
    )


@pytest.mark.parametrize('use_vertex', [True, False])
def test_t_schema_for_list_of_pydantic_schema(client):
  """Tests t_schema when list[pydantic.BaseModel] is passed to response_schema."""
  transformed_schema = _transformers.t_schema(client, list[CountryInfo])
  assert isinstance(transformed_schema, types.Schema)
  assert isinstance(transformed_schema.items, types.Schema)

  for schema_property in transformed_schema.items.properties:
    assert schema_property in country_info_fields
    assert isinstance(
        transformed_schema.items.properties[schema_property], types.Schema
    )


@pytest.mark.parametrize('use_vertex', [True, False])
def test_t_schema_for_null_fields(client):
  """Tests t_schema when null fields are present."""
  transformed_schema = _transformers.t_schema(client, CountryInfoWithNullFields)
  assert isinstance(transformed_schema, types.Schema)
  assert transformed_schema.properties['population'].nullable


#@pytest.mark.parametrize('use_vertex', [True, False])
def test_schema_with_no_null_fields_is_unchanged():
  """Tests handle_null_fields() doesn't change anything when no null fields are present."""
  test_properties = {
      'name': {'title': 'Name', 'type': 'string'},
      'total_area_sq_mi': {
          'anyOf': [{'type': 'integer'}, {'type': 'float'}],
          'default': 'null',
          'title': 'Total Area Sq Mi',
      },
  }

  for _, schema in test_properties.items():
    schema_before = copy.deepcopy(schema)
    _transformers.handle_null_fields(schema)
    assert schema_before == schema


@pytest.mark.parametrize('use_vertex', [True, False])
def test_schema_with_default_value(client):

  transformed_schema = _transformers.t_schema(
      client._api_client, CountryInfoWithDefaultValue
  )
  expected_schema = types.Schema(
      properties={
          'name': types.Schema(
              type='STRING',
              title='Name',
          ),
          'population': types.Schema(
              type='INTEGER',
              default=0,
              title='Population',
          ),
      },
      type='OBJECT',
      required=['name'],
      title='CountryInfoWithDefaultValue',
      property_ordering=['name', 'population'],
  )

  assert transformed_schema == expected_schema


@pytest.mark.parametrize('use_vertex', [True, False])
def test_schema_with_any_of(client):
  transformed_schema = _transformers.t_schema(client, CountryInfoWithAnyOf)
  expected_schema = types.Schema(
      properties={
          'name': types.Schema(
              type='STRING',
              title='Name',
          ),
          'restaurants_per_capita': types.Schema(
              any_of=[
                  types.Schema(type='INTEGER'),
                  types.Schema(type='NUMBER'),
              ],
              title='Restaurants Per Capita',
          ),
      },
      type='OBJECT',
      required=['name', 'restaurants_per_capita'],
      title='CountryInfoWithAnyOf',
      property_ordering=['name', 'restaurants_per_capita'],
  )

  assert transformed_schema == expected_schema


@pytest.mark.parametrize('use_vertex', [True, False])
def test_complex_dict_schema_with_anyof_is_unchanged(client):
  """When a dict schema is passed to process_schema, the only change should be camel-casing anyOf."""
  if client.vertexai:
    dict_schema = {
        'type': 'OBJECT',
        'title': 'Fruit Basket',
        'description': 'A structured representation of a fruit basket',
        'required': ['fruit'],
        'properties': {
            'fruit': {
                'type': 'ARRAY',
                'description': 'An ordered list of the fruit in the basket',
                'items': {
                    'description': 'A piece of fruit',
                    'anyOf': [
                        {
                            'title': 'Apple',
                            'description': 'Describes an apple',
                            'type': 'OBJECT',
                            'properties': {
                                'type': {
                                    'type': 'STRING',
                                    'description': "Always 'apple'",
                                },
                                'color': {
                                    'type': 'STRING',
                                    'description': (
                                        "The color of the apple (e.g., 'red')"
                                    ),
                                },
                            },
                            'propertyOrdering': ['type', 'color'],
                            'required': ['type', 'color'],
                        },
                        {
                            'title': 'Orange',
                            'description': 'Describes an orange',
                            'type': 'OBJECT',
                            'properties': {
                                'type': {
                                    'type': 'STRING',
                                    'description': "Always 'orange'",
                                },
                                'size': {
                                    'type': 'STRING',
                                    'description': (
                                        'The size of the orange (e.g.,'
                                        " 'medium')"
                                    ),
                                },
                            },
                            'propertyOrdering': ['type', 'size'],
                            'required': ['type', 'size'],
                        },
                    ],
                },
            }
        },
    }

    schema_before = copy.deepcopy(dict_schema)
    _transformers.process_schema(dict_schema, client)

    assert schema_before == dict_schema


@pytest.mark.parametrize('use_vertex', [True, False])
def test_process_schema_converts_const_to_enum(client):
  """The 'const' field should be converted to a singleton 'enum'."""
  schema = {
      'type': 'STRING',
      'const': 'FOO',
  }
  expected_schema = {
      'type': 'STRING',
      'enum': ['FOO'],
  }

  _transformers.process_schema(schema, client)

  assert schema == expected_schema


@pytest.mark.parametrize('use_vertex', [True, False])
def test_process_schema_forbids_non_string_const(client):
  """The 'const' field only works for strings."""
  schema = {
      'type': 'INTEGER',
      'const': 123,
  }

  with pytest.raises(ValueError, match='.*Literal values must be strings.*'):
    _transformers.process_schema(schema, client)


@pytest.mark.parametrize(
    'use_vertex,order_properties',
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_process_schema_order_properties_propagates_into_defs(
    client, order_properties
):
  """The `order_properties` setting should apply to '$defs'."""
  schema = {
      '$ref': '#/$defs/Foo',
      '$defs': {
          'Foo': {
              'type': 'OBJECT',
              'properties': {
                  'foo': {'type': 'STRING'},
                  'bar': {'type': 'STRING'},
              },
          },
      },
  }
  schema_without_property_ordering = {
      'type': 'OBJECT',
      'properties': {
          'foo': {'type': 'STRING'},
          'bar': {'type': 'STRING'},
      },
  }
  schema_with_property_ordering = {
      'type': 'OBJECT',
      'properties': {
          'foo': {'type': 'STRING'},
          'bar': {'type': 'STRING'},
      },
      'property_ordering': ['foo', 'bar'],
  }

  _transformers.process_schema(
      schema, client, order_properties=order_properties
  )

  if order_properties:
    assert schema == schema_with_property_ordering
  else:
    assert schema == schema_without_property_ordering


@pytest.mark.parametrize(
    'use_vertex,order_properties',
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_process_schema_order_properties_propagates_into_items(
    client, order_properties
):
  """The `order_properties` setting should apply to 'items'."""
  schema = {
      'type': 'ARRAY',
      'items': {
          'type': 'OBJECT',
          'properties': {
              'foo': {'type': 'STRING'},
              'bar': {'type': 'STRING'},
          },
      },
  }
  schema_without_property_ordering = copy.deepcopy(schema)
  schema_with_property_ordering = {
      'type': 'ARRAY',
      'items': {
          'type': 'OBJECT',
          'properties': {
              'foo': {'type': 'STRING'},
              'bar': {'type': 'STRING'},
          },
          'property_ordering': ['foo', 'bar'],
      },
  }

  _transformers.process_schema(
      schema, client, order_properties=order_properties
  )

  if order_properties:
    assert schema == schema_with_property_ordering
  else:
    assert schema == schema_without_property_ordering


@pytest.mark.parametrize(
    'use_vertex,order_properties',
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_process_schema_order_properties_propagates_into_prefix_items(
    client, order_properties
):
  """The `order_properties` setting should apply to 'prefixItems'."""
  schema = {
      'type': 'ARRAY',
      'prefixItems': [
          {
              'type': 'OBJECT',
              'properties': {
                  'foo': {'type': 'STRING'},
                  'bar': {'type': 'STRING'},
              },
          },
      ],
  }
  schema_without_property_ordering = copy.deepcopy(schema)
  schema_with_property_ordering = {
      'type': 'ARRAY',
      'prefixItems': [
          {
              'type': 'OBJECT',
              'properties': {
                  'foo': {'type': 'STRING'},
                  'bar': {'type': 'STRING'},
              },
              'property_ordering': ['foo', 'bar'],
          },
      ],
  }

  _transformers.process_schema(
      schema, client, order_properties=order_properties
  )

  if order_properties:
    assert schema == schema_with_property_ordering
  else:
    assert schema == schema_without_property_ordering


@pytest.mark.parametrize(
    'use_vertex,order_properties',
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_process_schema_order_properties_propagates_into_properties(
    client, order_properties
):
  """The `order_properties` setting should apply to 'properties'."""
  schema = {
      'type': 'OBJECT',
      'properties': {
          'xyz': {
              'type': 'OBJECT',
              'properties': {
                  'foo': {'type': 'STRING'},
                  'bar': {'type': 'STRING'},
              },
          },
          'abc': {'type': 'STRING'},
      },
  }
  schema_without_property_ordering = copy.deepcopy(schema)
  schema_with_property_ordering = {
      'type': 'OBJECT',
      'properties': {
          'xyz': {
              'type': 'OBJECT',
              'properties': {
                  'foo': {'type': 'STRING'},
                  'bar': {'type': 'STRING'},
              },
              'property_ordering': ['foo', 'bar'],
          },
          'abc': {'type': 'STRING'},
      },
      'property_ordering': ['xyz', 'abc'],
  }

  _transformers.process_schema(
      schema, client, order_properties=order_properties
  )

  if order_properties:
    assert schema == schema_with_property_ordering
  else:
    assert schema == schema_without_property_ordering


@pytest.mark.parametrize(
    'use_vertex,order_properties',
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_process_schema_order_properties_propagates_into_additional_properties(
    client, order_properties
):
  """The `order_properties` setting should apply to 'additionalProperties'."""
  schema = {
      'type': 'OBJECT',
      'additionalProperties': {
          'type': 'OBJECT',
          'properties': {
              'foo': {'type': 'STRING'},
              'bar': {'type': 'STRING'},
          },
      },
  }
  schema_without_property_ordering = copy.deepcopy(schema)
  schema_with_property_ordering = {
      'type': 'OBJECT',
      'additionalProperties': {
          'type': 'OBJECT',
          'properties': {
              'foo': {'type': 'STRING'},
              'bar': {'type': 'STRING'},
          },
          'property_ordering': ['foo', 'bar'],
      },
  }

  if client.vertexai:
    _transformers.process_schema(
        schema, client, order_properties=order_properties
    )

    if order_properties:
      assert schema == schema_with_property_ordering
    else:
      assert schema == schema_without_property_ordering
  else:
    with pytest.raises(ValueError) as e:
      _transformers.process_schema(
          schema, client, order_properties=order_properties
      )
    assert 'additionalProperties is not supported in the Gemini API.' in str(e)


@pytest.mark.parametrize(
    'use_vertex,order_properties',
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_process_schema_order_properties_propagates_into_any_of(
    client, order_properties
):
  """The `order_properties` setting should apply to 'anyOf'."""
  schema = {
      'anyOf': [
          {
              'type': 'OBJECT',
              'properties': {
                  'foo': {'type': 'STRING'},
                  'bar': {'type': 'STRING'},
              },
          },
          {'type': 'STRING'},
      ]
  }
  schema_without_property_ordering = copy.deepcopy(schema)
  schema_with_property_ordering = {
      'anyOf': [
          {
              'type': 'OBJECT',
              'properties': {
                  'foo': {'type': 'STRING'},
                  'bar': {'type': 'STRING'},
              },
              'property_ordering': ['foo', 'bar'],
          },
          {'type': 'STRING'},
      ]
  }

  _transformers.process_schema(
      schema, client, order_properties=order_properties
  )

  if order_properties:
    assert schema == schema_with_property_ordering
  else:
    assert schema == schema_without_property_ordering


@pytest.mark.parametrize('use_vertex', [True, False])
def test_t_schema_does_not_change_property_ordering_if_set(client):
  """Tests t_schema doesn't overwrite the property_ordering field if already set."""

  schema = CountryInfo.model_json_schema()
  custom_property_ordering = ['code', 'symbol', 'name']
  schema['property_ordering'] = custom_property_ordering.copy()

  transformed_schema = _transformers.t_schema(client, schema)
  assert transformed_schema.property_ordering == custom_property_ordering


@pytest.mark.parametrize('use_vertex', [True, False])
def test_t_schema_sets_property_ordering_for_json_schema(client):
  """Tests t_schema sets the property_ordering field for json schemas."""

  schema = CountryInfo.model_json_schema()

  transformed_schema = _transformers.t_schema(client, schema)
  assert transformed_schema.property_ordering == [
      'name',
      'population',
      'capital',
      'continent',
      'gdp',
      'official_language',
      'total_area_sq_mi',
  ]


@pytest.mark.parametrize('use_vertex', [True, False])
def test_t_schema_sets_property_ordering_for_schema_type(client):
  """Tests t_schema sets the property_ordering field for Schema types."""

  schema = types.Schema(
      properties={
          'name': types.Schema(
              type='STRING',
              title='Name',
          ),
          'population': types.Schema(
              type='INTEGER',
              default=0,
              title='Population',
          ),
      },
      type='OBJECT',
      required=['name'],
      title='CountryInfoWithDefaultValue',
  )

  transformed_schema = _transformers.t_schema(client, schema)
  assert transformed_schema.property_ordering == ['name', 'population']



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import logging
import pydantic

from ... import types


def _get_not_none_fields(model: pydantic.BaseModel) -> list[str]:
  """Returns field names in a Pydantic model whose values are not None."""
  return [
      field for field, value in model.model_dump().items() if value is not None
  ]


def test_empty_schema_conversion():
    """Test conversion of empty Schema to JSONSchema."""
    schema = types.Schema()
    json_schema = schema.json_schema
    not_none_field_names = _get_not_none_fields(json_schema)

    assert json_schema == types.JSONSchema()
    assert not_none_field_names == []


def test_not_null_type_conversion():
  """Test conversion of Schema.type to JSONSchema.type."""
  schema_types = [
      'OBJECT',
      'ARRAY',
      'STRING',
      'NUMBER',
      'BOOLEAN',
      'INTEGER',
  ]
  json_schema_types = [
      'object',
      'array',
      'string',
      'number',
      'boolean',
      'integer',
  ]
  for schema_type, expected_type in zip(schema_types, json_schema_types):
    schema = types.Schema(type=schema_type)
    json_schema = schema.json_schema
    not_none_field_names = _get_not_none_fields(json_schema)
    assert json_schema.type == types.JSONSchemaType(expected_type)
    assert not_none_field_names == ['type']


def test_unspecified_type_conversion():
  """Test conversion of Schema.type to JSONSchema.type."""
  schema = types.Schema(type='TYPE_UNSPECIFIED')
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.type is None
  assert not_none_field_names == []


def test_nullable_conversion():
  """Test conversion of Schema.nullable to JSONSchema.type."""
  schema = types.Schema(type='STRING', nullable=True)
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert set(json_schema.type) == set([
      types.JSONSchemaType('null'),
      types.JSONSchemaType('string')
  ])
  assert not_none_field_names == ['type']


def test_property_conversion():
  """Test conversion of Schema.properties to JSONSchema.properties."""
  schema = types.Schema(
      type='OBJECT',
      properties={
          'key1': types.Schema(type='STRING'),
          'key2': types.Schema(type='NUMBER'),
      },
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.properties == {
      'key1': types.JSONSchema(type=types.JSONSchemaType('string')),
      'key2': types.JSONSchema(type=types.JSONSchemaType('number')),
  }
  assert json_schema.type == types.JSONSchemaType('object')
  assert not_none_field_names == ['type', 'properties']


def test_complex_property_conversion():
  """Test conversion of complex Schema.properties to JSONSchema.properties."""
  schema = types.Schema(
      type='OBJECT',
      properties={
          'key1': types.Schema(
              type='OBJECT',
              properties={
                  'key2': types.Schema(type='STRING'),
                  'key3': types.Schema(type='NUMBER'),
              },
          ),
          'key2': types.Schema(type='ARRAY', items=types.Schema(type='STRING')),
      },
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.properties == {
      'key1': types.JSONSchema(
          type=types.JSONSchemaType('object'),
          properties={
              'key2': types.JSONSchema(type=types.JSONSchemaType('string')),
              'key3': types.JSONSchema(type=types.JSONSchemaType('number')),
          },
      ),
      'key2': types.JSONSchema(
          type=types.JSONSchemaType('array'),
          items=types.JSONSchema(type=types.JSONSchemaType('string')),
      ),
  }
  assert json_schema.type == types.JSONSchemaType('object')
  assert not_none_field_names == ['type', 'properties']


def test_items_conversion():
  """Test conversion of Schema.items to JSONSchema.items."""
  schema = types.Schema(
      type='ARRAY',
      items=types.Schema(type='STRING'),
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.type == types.JSONSchemaType('array')
  assert json_schema.items == types.JSONSchema(
      type=types.JSONSchemaType('string')
  )
  assert not_none_field_names == ['type', 'items']


def test_complex_items_conversion():
  """Test conversion of complex Schema.items to JSONSchema.items."""
  schema = types.Schema(
      type='ARRAY',
      items=types.Schema(
          type='OBJECT',
          properties={
              'key1': types.Schema(type='STRING'),
              'key2': types.Schema(type='NUMBER'),
          },
      ),
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.type == types.JSONSchemaType('array')
  assert json_schema.items == types.JSONSchema(
      type=types.JSONSchemaType('object'),
      properties={
          'key1': types.JSONSchema(type=types.JSONSchemaType('string')),
          'key2': types.JSONSchema(type=types.JSONSchemaType('number')),
      },
  )
  assert not_none_field_names == ['type', 'items']


def test_any_of_conversion():
  """Test conversion of Schema.any_of to JSONSchema.any_of."""
  schema = types.Schema(
      type='OBJECT',
      any_of=[
          types.Schema(type='STRING'),
          types.Schema(type='NUMBER'),
      ],
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.type == types.JSONSchemaType('object')
  assert json_schema.any_of == [
      types.JSONSchema(type=types.JSONSchemaType('string')),
      types.JSONSchema(type=types.JSONSchemaType('number')),
  ]
  assert not_none_field_names == ['type', 'any_of']


def test_complex_any_of_conversion():
  """Test conversion of complex Schema.any_of to JSONSchema.any_of."""
  schema = types.Schema(
      type='OBJECT',
      any_of=[
          types.Schema(
              type='OBJECT',
              properties={
                  'key1': types.Schema(type='STRING'),
                  'key2': types.Schema(type='NUMBER'),
              },
          ),
          types.Schema(type='ARRAY', items=types.Schema(type='STRING')),
      ],
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.type == types.JSONSchemaType('object')
  assert json_schema.any_of == [
      types.JSONSchema(
          type=types.JSONSchemaType('object'),
          properties={
              'key1': types.JSONSchema(type=types.JSONSchemaType('string')),
              'key2': types.JSONSchema(type=types.JSONSchemaType('number')),
          },
      ),
      types.JSONSchema(
          type=types.JSONSchemaType('array'),
          items=types.JSONSchema(type=types.JSONSchemaType('string')),
      ),
  ]
  assert not_none_field_names == ['type', 'any_of']


def test_example_conversion():
  """Test conversion of Schema.direct to JSONSchema.direct."""
  schema = types.Schema(
      example='this is an example',
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert not_none_field_names == []


def test_property_ordering_conversion():
  """Test conversion of Schema.property_ordering to JSONSchema.property_ordering."""
  schema = types.Schema(
      property_ordering=['a', 'b'],
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert not_none_field_names == []


def test_direct_conversion():
  """Test Schema fiedls that do not need to be converted."""
  schema = types.Schema(
      pattern='^[a-z]+$',
      default=1,
      max_length=10,
      title='title',
      min_length=2,
      min_properties=3,
      max_properties=7,
      description='description',
      enum=['enum1', 'enum2'],
      format='email',
      max_items=199,
      maximum=300,
      min_items=6,
      minimum=40,
      required=['required1', 'required2'],
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.pattern == '^[a-z]+$'
  assert json_schema.default == 1
  assert json_schema.max_length == 10
  assert json_schema.title == 'title'
  assert json_schema.min_length == 2
  assert json_schema.min_properties == 3
  assert json_schema.max_properties == 7
  assert json_schema.description == 'description'
  assert json_schema.enum == ['enum1', 'enum2']
  assert json_schema.format == 'email'
  assert json_schema.max_items == 199
  assert json_schema.maximum == 300
  assert json_schema.min_items == 6
  assert json_schema.minimum == 40
  assert json_schema.required == ['required1', 'required2']
  assert not_none_field_names.sort() == [
      'pattern',
      'default',
      'max_length',
      'title',
      'min_length',
      'min_properties',
      'max_properties',
      'description',
      'enum',
      'format',
      'max_items',
      'maximum',
      'min_items',
      'minimum',
      'required',
  ].sort()


def test_complex_any_of_conversion():
  schema = types.Schema(
      type=types.Type.OBJECT,
      title='Fruit Basket',
      description='A structured representation of a fruit basket',
      properties={
          'fruit': types.Schema(
              type=types.Type.ARRAY,
              description='An ordered list of the fruit in the basket',
              items=types.Schema(
                  any_of=[
                      types.Schema(
                          title='Apple',
                          description='Describes an apple',
                          type=types.Type.OBJECT,
                          properties={
                              'type': types.Schema(
                                  type=types.Type.STRING,
                                  description='Always "apple"',
                              ),
                              'variety': types.Schema(
                                  type=types.Type.STRING,
                                  description=(
                                      'The variety of apple (e.g., "Granny'
                                      ' Smith")'
                                  ),
                              ),
                          },
                          property_ordering=['type', 'variety'],
                          required=['type', 'variety'],
                      ),
                      types.Schema(
                          title='Orange',
                          description='Describes an orange',
                          type=types.Type.OBJECT,
                          properties={
                              'type': types.Schema(
                                  type=types.Type.STRING,
                                  description='Always "orange"',
                              ),
                              'variety': types.Schema(
                                  type=types.Type.STRING,
                                  description=(
                                      'The variety of orange (e.g.,"Navel'
                                      ' orange")'
                                  ),
                              ),
                          },
                          property_ordering=['type', 'variety'],
                          required=['type', 'variety'],
                      ),
                  ],
              ),
          ),
      },
      required=['fruit'],
  )
  json_schema = schema.json_schema
  not_none_field_names = _get_not_none_fields(json_schema)

  assert json_schema.type == types.JSONSchemaType('object')
  assert json_schema.title == 'Fruit Basket'
  assert json_schema.description == 'A structured representation of a fruit basket'
  assert json_schema.properties == {
      'fruit': types.JSONSchema(
          type=types.JSONSchemaType('array'),
          description='An ordered list of the fruit in the basket',
          items=types.JSONSchema(
              any_of=[
                  types.JSONSchema(
                      title='Apple',
                      description='Describes an apple',
                      type=types.JSONSchemaType('object'),
                      properties={
                          'type': types.JSONSchema(
                              type=types.JSONSchemaType('string'),
                              description='Always "apple"',
                          ),
                          'variety': types.JSONSchema(
                              type=types.JSONSchemaType('string'),
                              description=(
                                  'The variety of apple (e.g., "Granny'
                                  ' Smith")'
                              ),
                          ),
                      },
                      required=['type', 'variety'],
                  ),
                  types.JSONSchema(
                      title='Orange',
                      description='Describes an orange',
                      type=types.JSONSchemaType('object'),
                      properties={
                          'type': types.JSONSchema(
                              type=types.JSONSchemaType('string'),
                              description='Always "orange"',
                          ),
                          'variety': types.JSONSchema(
                              type=types.JSONSchemaType('string'),
                              description=(
                                  'The variety of orange (e.g.,"Navel orange")'
                              ),
                          ),
                      },
                      required=['type', 'variety'],
                  ),
              ],
          ),
      ),
  }
  assert json_schema.required == ['fruit']
  assert not_none_field_names == [
      'type',
      'title',
      'description',
      'properties',
      'required',
  ]


def test_json_schema_logs_only_once(caplog):
  """Test that the info message is logged only once across multiple json_schema calls."""
  from ... import types as types_module

  types_module._json_schema_warning_logged = False

  caplog.set_level(logging.INFO, logger='google_genai.types')

  schema1 = types_module.Schema(type='STRING')
  json_schema1 = schema1.json_schema

  assert len(caplog.records) == 1
  assert 'Json Schema is now supported natively' in caplog.text
  assert 'response_json_schema' in caplog.text

  schema2 = types_module.Schema(type='NUMBER')
  json_schema2 = schema2.json_schema

  assert len(caplog.records) == 1

  schema3 = types_module.Schema(type='OBJECT')
  json_schema3 = schema3.json_schema

  assert len(caplog.records) == 1

  assert json_schema1.type == types_module.JSONSchemaType('string')
  assert json_schema2.type == types_module.JSONSchemaType('number')
  assert json_schema3.type == types_module.JSONSchemaType('object')

  types_module._json_schema_warning_logged = False



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Code generated by the Google Gen AI SDK generator DO NOT EDIT.

import json
import logging
from typing import Any, Optional, TypeVar, Union
from urllib.parse import urlencode

from . import _api_module
from . import _common
from . import types
from ._common import get_value_by_path as getv
from ._common import set_value_by_path as setv


logger = logging.getLogger('google_genai.operations')


def _FetchPredictOperationParameters_to_vertex(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
  to_object: dict[str, Any] = {}
  if getv(from_object, ['operation_name']) is not None:
    setv(to_object, ['operationName'], getv(from_object, ['operation_name']))

  if getv(from_object, ['resource_name']) is not None:
    setv(
        to_object,
        ['_url', 'resourceName'],
        getv(from_object, ['resource_name']),
    )

  return to_object


def _GetOperationParameters_to_mldev(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
  to_object: dict[str, Any] = {}
  if getv(from_object, ['operation_name']) is not None:
    setv(
        to_object,
        ['_url', 'operationName'],
        getv(from_object, ['operation_name']),
    )

  return to_object


def _GetOperationParameters_to_vertex(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
  to_object: dict[str, Any] = {}
  if getv(from_object, ['operation_name']) is not None:
    setv(
        to_object,
        ['_url', 'operationName'],
        getv(from_object, ['operation_name']),
    )

  return to_object


def _GetProjectOperationParameters_to_vertex(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
  to_object: dict[str, Any] = {}
  if getv(from_object, ['operation_id']) is not None:
    setv(
        to_object, ['_url', 'operation_id'], getv(from_object, ['operation_id'])
    )

  return to_object


class Operations(_api_module.BaseModule):

  def _get_videos_operation(
      self,
      *,
      operation_name: str,
      config: Optional[types.GetOperationConfigOrDict] = None,
  ) -> Any:
    parameter_model = types._GetOperationParameters(
        operation_name=operation_name,
        config=config,
    )

    request_url_dict: Optional[dict[str, str]]

    if self._api_client.vertexai:
      request_dict = _GetOperationParameters_to_vertex(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = '{operationName}'.format_map(request_url_dict)
      else:
        path = '{operationName}'
    else:
      request_dict = _GetOperationParameters_to_mldev(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = '{operationName}'.format_map(request_url_dict)
      else:
        path = '{operationName}'
    query_params = request_dict.get('_query')
    if query_params:
      path = f'{path}?{urlencode(query_params)}'
    # TODO: remove the hack that pops config.
    request_dict.pop('config', None)

    http_options: Optional[types.HttpOptions] = None
    if (
        parameter_model.config is not None
        and parameter_model.config.http_options is not None
    ):
      http_options = parameter_model.config.http_options

    request_dict = _common.convert_to_dict(request_dict)
    request_dict = _common.encode_unserializable_types(request_dict)

    response = self._api_client.request('get', path, request_dict, http_options)

    response_dict = {} if not response.body else json.loads(response.body)

    return response_dict

  def _fetch_predict_videos_operation(
      self,
      *,
      operation_name: str,
      resource_name: str,
      config: Optional[types.FetchPredictOperationConfigOrDict] = None,
  ) -> Any:
    parameter_model = types._FetchPredictOperationParameters(
        operation_name=operation_name,
        resource_name=resource_name,
        config=config,
    )

    request_url_dict: Optional[dict[str, str]]
    if not self._api_client.vertexai:
      raise ValueError('This method is only supported in the Vertex AI client.')
    else:
      request_dict = _FetchPredictOperationParameters_to_vertex(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = '{resourceName}:fetchPredictOperation'.format_map(
            request_url_dict
        )
      else:
        path = '{resourceName}:fetchPredictOperation'

    query_params = request_dict.get('_query')
    if query_params:
      path = f'{path}?{urlencode(query_params)}'
    # TODO: remove the hack that pops config.
    request_dict.pop('config', None)

    http_options: Optional[types.HttpOptions] = None
    if (
        parameter_model.config is not None
        and parameter_model.config.http_options is not None
    ):
      http_options = parameter_model.config.http_options

    request_dict = _common.convert_to_dict(request_dict)
    request_dict = _common.encode_unserializable_types(request_dict)

    response = self._api_client.request(
        'post', path, request_dict, http_options
    )

    response_dict = {} if not response.body else json.loads(response.body)

    return response_dict

  def _get(
      self,
      *,
      operation_id: str,
      config: Optional[types.GetOperationConfigOrDict] = None,
  ) -> types.ProjectOperation:
    parameter_model = types._GetProjectOperationParameters(
        operation_id=operation_id,
        config=config,
    )

    request_url_dict: Optional[dict[str, str]]
    if not self._api_client.vertexai:
      raise ValueError('This method is only supported in the Vertex AI client.')
    else:
      request_dict = _GetProjectOperationParameters_to_vertex(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = 'operations/{operation_id}'.format_map(request_url_dict)
      else:
        path = 'operations/{operation_id}'

    query_params = request_dict.get('_query')
    if query_params:
      path = f'{path}?{urlencode(query_params)}'
    # TODO: remove the hack that pops config.
    request_dict.pop('config', None)

    http_options: Optional[types.HttpOptions] = None
    if (
        parameter_model.config is not None
        and parameter_model.config.http_options is not None
    ):
      http_options = parameter_model.config.http_options

    request_dict = _common.convert_to_dict(request_dict)
    request_dict = _common.encode_unserializable_types(request_dict)

    response = self._api_client.request('get', path, request_dict, http_options)

    response_dict = {} if not response.body else json.loads(response.body)

    return_value = types.ProjectOperation._from_response(
        response=response_dict, kwargs=parameter_model.model_dump()
    )

    self._api_client._verify_response(return_value)
    return return_value

  T = TypeVar('T', bound=types.Operation)

  def get(
      self,
      operation: T,
      *,
      config: Optional[types.GetOperationConfigOrDict] = None,
  ) -> T:
    """Gets the status of a long-running operation.

    Args:
      operation (Operation): The operation instance to get the status for.
      config (GetOperationConfig): Configuration for getting the operation.

    Returns:
      Operation: The updated operation instance with the latest status or
      result.

    Usage:

    .. code-block:: python

      import time

      operation = client.models.generate_videos(
          model="veo-2.0-generate-001",
          source=types.GenerateVideosSource(
              prompt="A neon hologram of a cat driving at top speed",
          ),
      )
      while not operation.done:
          time.sleep(10)
          operation = client.operations.get(operation)

      print(operation.result)
    """
    # Currently, only GenerateVideosOperation is supported.
    # TODO(b/398040607): Support short form names
    operation_name = operation.name
    if not operation_name:
      raise ValueError('Operation name is empty.')

    # TODO(b/398233524): Cast operation types
    if self._api_client.vertexai:
      resource_name = operation_name.rpartition('/operations/')[0]
      http_options = types.HttpOptions()
      if isinstance(config, dict):
        dict_options = config.get('http_options', None)
        if dict_options is not None:
          http_options = types.HttpOptions(**dict(dict_options))
      elif isinstance(config, types.GetOperationConfig) and config is not None:
        http_options = (
            config.http_options
            if config.http_options is not None
            else types.HttpOptions()
        )
      fetch_operation_config = types.FetchPredictOperationConfig(
          http_options=http_options
      )
      response_dict = self._fetch_predict_videos_operation(
          operation_name=operation_name,
          resource_name=resource_name,
          config=fetch_operation_config,
      )

      response_operation = operation.from_api_response(
          response_dict, is_vertex_ai=True
      )

      self._api_client._verify_response(response_operation)  # type: ignore [arg-type]
      return response_operation  # type: ignore[no-any-return]
    else:
      response_dict = self._get_videos_operation(
          operation_name=operation_name,
          config=config,
      )
      response_operation = operation.from_api_response(
          response_dict, is_vertex_ai=False
      )

      self._api_client._verify_response(response_operation)  # type: ignore [arg-type]
      return response_operation  # type: ignore[no-any-return]


class AsyncOperations(_api_module.BaseModule):

  async def _get_videos_operation(
      self,
      *,
      operation_name: str,
      config: Optional[types.GetOperationConfigOrDict] = None,
  ) -> Any:
    parameter_model = types._GetOperationParameters(
        operation_name=operation_name,
        config=config,
    )

    request_url_dict: Optional[dict[str, str]]

    if self._api_client.vertexai:
      request_dict = _GetOperationParameters_to_vertex(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = '{operationName}'.format_map(request_url_dict)
      else:
        path = '{operationName}'
    else:
      request_dict = _GetOperationParameters_to_mldev(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = '{operationName}'.format_map(request_url_dict)
      else:
        path = '{operationName}'
    query_params = request_dict.get('_query')
    if query_params:
      path = f'{path}?{urlencode(query_params)}'
    # TODO: remove the hack that pops config.
    request_dict.pop('config', None)

    http_options: Optional[types.HttpOptions] = None
    if (
        parameter_model.config is not None
        and parameter_model.config.http_options is not None
    ):
      http_options = parameter_model.config.http_options

    request_dict = _common.convert_to_dict(request_dict)
    request_dict = _common.encode_unserializable_types(request_dict)

    response = await self._api_client.async_request(
        'get', path, request_dict, http_options
    )

    response_dict = {} if not response.body else json.loads(response.body)

    return response_dict

  async def _fetch_predict_videos_operation(
      self,
      *,
      operation_name: str,
      resource_name: str,
      config: Optional[types.FetchPredictOperationConfigOrDict] = None,
  ) -> Any:
    parameter_model = types._FetchPredictOperationParameters(
        operation_name=operation_name,
        resource_name=resource_name,
        config=config,
    )

    request_url_dict: Optional[dict[str, str]]
    if not self._api_client.vertexai:
      raise ValueError('This method is only supported in the Vertex AI client.')
    else:
      request_dict = _FetchPredictOperationParameters_to_vertex(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = '{resourceName}:fetchPredictOperation'.format_map(
            request_url_dict
        )
      else:
        path = '{resourceName}:fetchPredictOperation'

    query_params = request_dict.get('_query')
    if query_params:
      path = f'{path}?{urlencode(query_params)}'
    # TODO: remove the hack that pops config.
    request_dict.pop('config', None)

    http_options: Optional[types.HttpOptions] = None
    if (
        parameter_model.config is not None
        and parameter_model.config.http_options is not None
    ):
      http_options = parameter_model.config.http_options

    request_dict = _common.convert_to_dict(request_dict)
    request_dict = _common.encode_unserializable_types(request_dict)

    response = await self._api_client.async_request(
        'post', path, request_dict, http_options
    )

    response_dict = {} if not response.body else json.loads(response.body)

    return response_dict

  async def _get(
      self,
      *,
      operation_id: str,
      config: Optional[types.GetOperationConfigOrDict] = None,
  ) -> types.ProjectOperation:
    parameter_model = types._GetProjectOperationParameters(
        operation_id=operation_id,
        config=config,
    )

    request_url_dict: Optional[dict[str, str]]
    if not self._api_client.vertexai:
      raise ValueError('This method is only supported in the Vertex AI client.')
    else:
      request_dict = _GetProjectOperationParameters_to_vertex(parameter_model)
      request_url_dict = request_dict.get('_url')
      if request_url_dict:
        path = 'operations/{operation_id}'.format_map(request_url_dict)
      else:
        path = 'operations/{operation_id}'

    query_params = request_dict.get('_query')
    if query_params:
      path = f'{path}?{urlencode(query_params)}'
    # TODO: remove the hack that pops config.
    request_dict.pop('config', None)

    http_options: Optional[types.HttpOptions] = None
    if (
        parameter_model.config is not None
        and parameter_model.config.http_options is not None
    ):
      http_options = parameter_model.config.http_options

    request_dict = _common.convert_to_dict(request_dict)
    request_dict = _common.encode_unserializable_types(request_dict)

    response = await self._api_client.async_request(
        'get', path, request_dict, http_options
    )

    response_dict = {} if not response.body else json.loads(response.body)

    return_value = types.ProjectOperation._from_response(
        response=response_dict, kwargs=parameter_model.model_dump()
    )

    self._api_client._verify_response(return_value)
    return return_value

  T = TypeVar('T', bound=types.Operation)

  async def get(
      self,
      operation: T,
      *,
      config: Optional[types.GetOperationConfigOrDict] = None,
  ) -> T:
    """Gets the status of a long-running operation.

    Args:
      operation (Operation): The operation instance to get the status for.
      config (GetOperationConfig): Configuration for getting the operation.

    Returns:
      Operation: The updated operation instance with the latest status or
      result.

    Usage:

    .. code-block:: python

      import asyncio

      operation = await client.aio.models.generate_videos(
          model="veo-2.0-generate-001",
          source=types.GenerateVideosSource(
              prompt="A neon hologram of a cat driving at top speed",
          ),
      )
      while not operation.done:
          await asyncio.sleep(10)
          operation = await client.aio.operations.get(operation)

      print(operation.result)
    """
    # Currently, only GenerateVideosOperation is supported.
    operation_name = operation.name
    if not operation_name:
      raise ValueError('Operation name is empty.')

    if self._api_client.vertexai:
      resource_name = operation_name.rpartition('/operations/')[0]
      http_options = types.HttpOptions()
      if isinstance(config, dict):
        dict_options = config.get('http_options', None)
        if dict_options is not None:
          http_options = types.HttpOptions(**dict(dict_options))
      elif isinstance(config, types.GetOperationConfig) and config is not None:
        http_options = (
            config.http_options
            if config.http_options is not None
            else types.HttpOptions()
        )
      fetch_operation_config = types.FetchPredictOperationConfig(
          http_options=http_options
      )
      response_dict = await self._fetch_predict_videos_operation(
          operation_name=operation_name,
          resource_name=resource_name,
          config=fetch_operation_config,
      )
      response_operation = operation.from_api_response(
          response_dict, is_vertex_ai=True
      )
      return response_operation  # type: ignore[no-any-return]
    else:
      response_dict = await self._get_videos_operation(
          operation_name=operation_name,
          config=config,
      )
      response_operation = operation.from_api_response(
          response_dict, is_vertex_ai=False
      )
      return response_operation  # type: ignore[no-any-return]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for batches._create_embeddings()"""

import pytest

from ... import types
from .. import pytest_helper

_MLDEV_EMBEDDING_BATCH_INLINE_OPERATION_NAME = (
    'batches/wdx71o8cgbzoa6gg3be1mg7g8ulrhapcjgo3'
)
_MLDEV_EMBEDDING_BATCH_FILE_OPERATION_NAME = (
    'batches/507oatd242het8ox60pwsmn7tcmtkrj8itff'
)

_DISPLAY_NAME = 'test_batch'
_MLDEV_EMBEDDING_MODEL = 'gemini-embedding-001'
_EMBED_CONTENT_FILE_NAME = 'files/mq9e3mg3u2y5'
_INLINED_EMBED_CONTENT_REQUESTS = {
    'config': {'output_dimensionality': 64},
    'contents': [
        {
            'parts': [{
                'text': '1',
            }],
        },
        {
            'parts': [{
                'text': '2',
            }],
        },
        {
            'parts': [{
                'text': '3',
            }],
        },
    ],
}

test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name='test_from_inlined',
        parameters=types._CreateEmbeddingsBatchJobParameters(
            model=_MLDEV_EMBEDDING_MODEL,
            src={'inlined_requests': _INLINED_EMBED_CONTENT_REQUESTS},
            config={
                'display_name': _DISPLAY_NAME,
            },
        ),
        exception_if_vertex='Vertex AI does not support',
    ),
]

pytestmark = [
    pytest.mark.usefixtures('mock_timestamped_unique_name'),
    pytest_helper.setup(
        file=__file__,
        globals_for_file=globals(),
        test_method='batches.create_embeddings',
        test_table=test_table,
        http_options={
            'api_version': 'v1alpha',
            'base_url': (
                'https://autopush-generativelanguage.sandbox.googleapis.com'
            ),
        },
    ),
]


@pytest.mark.asyncio
async def test_async_from_inline(client):
  with pytest_helper.exception_if_vertex(client, ValueError):
    batch_job = await client.aio.batches.create_embeddings(
        model=_MLDEV_EMBEDDING_MODEL,
        src={'inlined_requests': _INLINED_EMBED_CONTENT_REQUESTS},
    )
    assert batch_job.name.startswith('batches/')


def test_from_file(client):
  """Tests creating a batch job with an embedding file name."""
  with pytest_helper.exception_if_vertex(client, ValueError):
    batch_job = client.batches.create_embeddings(
        model=_MLDEV_EMBEDDING_MODEL,
        src={'file_name': _EMBED_CONTENT_FILE_NAME},
        config={
            'display_name': _DISPLAY_NAME,
        },
    )
    assert batch_job.name.startswith('batches/')
    assert batch_job.model == 'models/' + _MLDEV_EMBEDDING_MODEL


@pytest.mark.asyncio
async def test_async_from_file(client):
  with pytest_helper.exception_if_vertex(client, ValueError):
    batch_job = await client.aio.batches.create_embeddings(
        model=_MLDEV_EMBEDDING_MODEL,
        src={'file_name': _EMBED_CONTENT_FILE_NAME},
        config={
            'display_name': _DISPLAY_NAME,
        },
    )
    assert batch_job.name.startswith('batches/')
    assert (
        batch_job.model == 'models/' + _MLDEV_EMBEDDING_MODEL
    )  # Converted to Gemini full name.


def test_get_inline(client):
  """Tests getting a batch job that used inline requests."""
  with pytest_helper.exception_if_vertex(client, ValueError):
    name = _MLDEV_EMBEDDING_BATCH_INLINE_OPERATION_NAME
    batch_job = client.batches.get(name=name)
    assert batch_job.dest.inlined_embed_content_responses is not None


@pytest.mark.asyncio
async def test_async_get_inline(client):
  with pytest_helper.exception_if_vertex(client, ValueError):
    name = _MLDEV_EMBEDDING_BATCH_INLINE_OPERATION_NAME
    batch_job = await client.aio.batches.get(name=name)

    assert batch_job.dest.inlined_embed_content_responses is not None


def test_get_file(client):
  """Tests getting a batch job that used a file source."""
  with pytest_helper.exception_if_vertex(client, ValueError):
    name = _MLDEV_EMBEDDING_BATCH_FILE_OPERATION_NAME
    batch_job = client.batches.get(name=name)
    assert batch_job.dest.file_name is not None


@pytest.mark.asyncio
async def test_async_get_file(client):
  with pytest_helper.exception_if_vertex(client, ValueError):
    name = _MLDEV_EMBEDDING_BATCH_FILE_OPERATION_NAME
    batch_job = await client.aio.batches.get(name=name)

    assert batch_job.dest.file_name is not None
